import click

from gcs_cli.helpers import (
    load_certificate,
    load_certificate_chain,
    load_private_key,
    record_print_result_message,
    requires_login,
)
from gcs_cli.parsing import domain_update_options, globus_command
from gcs_cli.services.gcs import get_client


@globus_command(
    "update",
    short_help="Update the custom domain and certificates for an endpoint",
    disable_options=["format"],
)
@domain_update_options(endpoint=True)
@requires_login
def domain_update(
    certificate_path=None,
    certificate_chain_path=None,
    private_key_path=None,
    managed=None,
    wildcard=None,
    domain=None,
):
    """
    Update the domain to use for an endpoint. This can set either a wildcard
    or a non-wildcard domain. If a wildcard domain is used, then all collections
    on this endpoint will be made subdomains of this domain. If not, only the
    endpoint domain is changed.

    The certificate, chain, and key can either be managed or unmanaged. If
    managed, the data is encrypted and shared between data transfer nodes
    serving this endpoint. If unmanaged, the paths specified on the
    command-line must exist on each data transfer node, and Globus Connect
    Server will only share the location for those files between data transfer
    nodes.
    """
    client = get_client()

    doc = {
        "DATA_TYPE": "domain#1.0.0",
        "wildcard": wildcard,
        "domain_name": domain,
    }
    if not managed:
        doc["certificate_path"] = certificate_path
        if certificate_chain_path:
            doc["certificate_chain_path"] = certificate_chain_path
        doc["private_key_path"] = private_key_path
    else:
        doc["certificate"] = load_certificate(certificate_path)

        if certificate_chain_path:
            doc["certificate_chain"] = load_certificate_chain(certificate_chain_path)

        def get_passphrase():
            click.prompt("Enter Passphrase: ", hide_input=True)

        doc["private_key"] = load_private_key(
            private_key_path, passphrase=get_passphrase
        )

    res = client.put_endpoint_domain(doc)
    record_print_result_message(res)
