import errno
import json
import socket
import ssl
import textwrap
import traceback
import urllib.parse

import click
import requests
import urllib3.exceptions

import globus_sdk
from globus_sdk import GCSAPIError

from gcs_cli.parsing.stateful_options import get_show_stack_trace, outformat_is_text
from gcs_cli.services.auth import get_client


def handle_excinfo(excinfo):
    exception_type, exception, _ = excinfo

    if get_show_stack_trace():
        click.echo("".join(traceback.format_exception(*excinfo)), err=True)
    if isinstance(
        exception, (click.ClickException, click.exceptions.Abort, click.exceptions.Exit)
    ):
        raise exception
    elif isinstance(exception, GCSAPIError):
        _gcsapi_error_hook(exception)
    elif isinstance(exception, globus_sdk.GlobusAPIError):
        _default_api_error_hook(exception)
    elif isinstance(exception, globus_sdk.GlobusConnectionError):
        _default_connection_error_hook(exception)
    else:
        click.echo(u"{}: {}".format(exception_type.__name__, exception))
    click.get_current_context().exit(1)


def _default_connection_error_hook(exception):
    click.secho(
        "Error contacting the GCS Manager API", err=True, color="red", bold=True
    )

    requests_exc = exception.underlying_exception
    netloc = None
    if requests_exc and isinstance(requests_exc, requests.exceptions.ConnectionError):
        try:
            request_url = requests_exc.request.url
            netloc = urllib.parse.urlparse(request_url).netloc
        except Exception:
            pass

        for arg in requests_exc.args:
            if isinstance(arg, urllib3.exceptions.RequestError):
                context = getattr(arg, "__context__", None)
                while context:
                    if isinstance(context, (socket.gaierror, ConnectionError)):
                        _handle_network_error(netloc, context)
                        context = None
                    elif isinstance(context, ssl.SSLError):
                        _handle_ssl_error(netloc, context)
                        context = None
                    else:
                        context = getattr(context, "__context__", None)


def _handle_network_error(netloc, exception):
    if exception.errno == socket.EAI_NONAME:
        click.secho(
            "\n".join(
                ["Error resolving {}".format(netloc)]
                + textwrap.wrap(
                    "This may be because the endpoint is deleted, it is not deployed "
                    "on any nodes, or your DNS resolver is misconfigured."
                )
            ),
            err=True,
        )
    elif exception.errno == errno.ECONNREFUSED:
        click.secho(
            "\n".join(
                ["Error connecting to {}".format(netloc)]
                + textwrap.wrap(
                    "This may be because the GCS services are not running on the "
                    "endpoint nodes, or an endpoint node's IP address has changed."
                )
            ),
            err=True,
        )
    else:
        click.secho(
            "\n".join(
                textwrap.wrap(
                    "Error connecting to {}: {}".format(netloc, str(exception))
                )
            ),
            err=True,
        )


def _handle_ssl_error(netloc, exception):
    if exception.reason == "CERTIFICATE_VERIFY_FAILED":
        click.secho(
            "\n".join(
                ["SSL error connecting to {}: {}".format(netloc, exception.reason)]
                + textwrap.wrap(
                    "This may be because the trust roots for your system don't support "
                    "the Let's Encrypt CA or another virtual host is configured to use "
                    "the same server name as the GCS Manager."
                )
            ),
            err=True,
        )
    else:
        click.secho(
            "SSL error connecting to {}: {}".format(netloc, exception.reason),
            err=True,
        )


def _default_api_error_hook(exception):
    click.secho("An API Error Occurred", err=True, color="red", bold=True)
    click.secho(
        "HTTP status: {}".format(exception.http_status), err=True, color="yellow"
    )
    click.echo("code:        {}".format(exception.code), err=True)
    if exception.message:
        click.echo("message:     {}".format(exception.message), err=True)


def _auth_timeout_error_hook(exception):
    ids = exception.detail["identities"]
    if len(ids) > 1:
        cmdmsg = "To resolve, you should run one of these commands"
    else:
        cmdmsg = "To resolve, you should run this command"

    click.secho("An authentication timeout occurred", err=True, color="red", bold=True)
    click.echo(
        "This means you would have been authorized based on your identities, "
        "but you haven't logged in recently enough.",
        err=True,
    )
    click.secho(cmdmsg + " to update your authentication:", err=True, color="yellow")
    mfa_flag = ""
    if exception.detail.get("require_mfa", False):
        mfa_flag = "--mfa "
    for n in _try_resolve_ids_to_usernames(ids):
        click.echo(
            "  globus-connect-server session update {}{}".format(mfa_flag, n), err=True
        )


def _required_scopes_error_hook(exception):
    scopes = exception.detail["required_scopes"]
    click.secho(
        "This operation requires additional authentication.", err=True, color="yellow"
    )
    click.secho("Update your consents with:\n", err=True)
    click.echo(
        "  globus-connect-server session consent {}".format(" ".join(scopes)), err=True
    )


def _json_api_error_hook(exception):
    response = exception._underlying_response
    if exception.json_data:
        body = exception.json_data
    else:
        body = response.text
    headers = dict(response.headers)
    gcs_headers = {k: v for k, v in headers.items() if k.startswith("x-gcs-")}
    other_headers = {k: v for k, v in headers.items() if k not in gcs_headers}
    click.secho(
        json.dumps(
            {
                "is_error": True,
                "http_status": exception.http_status,
                "code": exception.code,
                "message": exception.message,
                "body": body,
                "gcs_headers": gcs_headers,
                "headers": other_headers,
            },
            indent=2,
            separators=(",", ": "),
            sort_keys=True,
        ),
        color="yellow",
    )


def _gcsapi_error_hook(exception):
    if outformat_is_text():
        if exception.detail_data_type == "authentication_timeout#1.1.0":
            _auth_timeout_error_hook(exception)
        elif exception.detail_data_type == "missing_required_scopes#1.0.0":
            _required_scopes_error_hook(exception)
        else:
            _default_api_error_hook(exception)
    else:
        _json_api_error_hook(exception)


def _try_resolve_ids_to_usernames(ids):
    """resolve IDs to usernames, falling back to the IDs which were given"""
    try:
        idmap = globus_sdk.IdentityMap(get_client())
        # add all to do batched lookups
        for i in ids:
            idmap.add(i)
        # do the lookups and produce the usernames
        return [idmap[i]["username"] for i in ids]
    # very conservative error handling -- if anything goes wrong, just return the
    # original IDs
    except Exception:
        return ids
