import sys

import click

from gcs_cli.error_handler import handle_excinfo
from gcs_cli.parsing.stateful_options import (
    explicit_host_option,
    format_option,
    show_stack_trace_option,
)
from gcs_cli.version import __version__ as version


def _common_options(f, disable_opts):
    if "format" not in disable_opts:
        f = format_option(f)
    if "version" not in disable_opts:
        f = click.version_option(version)(f)
    if "help" not in disable_opts:
        f = click.help_option("-h", "--help")(f)
    if "explicit_host" not in disable_opts:
        f = explicit_host_option(f)
    if "show_stack_trace" not in disable_opts:
        f = show_stack_trace_option(f)
    return f


def globus_command(*args, **kwargs):
    """
    A helper for decorating commands a-la `click.command`, but pulling the help string
    from `<function>.__doc__` by default. Fancy!
    """
    disable_opts = kwargs.pop("disable_options", [])

    def _inner_decorator(func):
        if "help" not in kwargs:
            kwargs["help"] = func.__doc__

        return _common_options(click.command(*args, **kwargs)(func), disable_opts)

    return _inner_decorator


class TopLevelGroup(click.Group):
    def invoke(self, ctx):
        try:
            return super().invoke(ctx)
        except Exception:
            handle_excinfo(sys.exc_info())


class CollectionCommandGroup(click.Group):
    """
    A custom command group which will alias `create-mapped` as `create`

    This lets the commands

      gcs collection create ...
      gcs collection create-mapped ...

    be treated nearly identically.

    create-guest is parallel to create-mapped, so the initial version of create-mapped
    allowed this aliasing for backwards compatibility. At present, `create-guest` is
    hidden, however, so we want to treat `create` as the primary name and
    `create-mapped` as the alias.

    If a future version starts using `create-mapped` and we update docs, older versions
    with this aliasing will continue to support those commands.
    """

    def get_command(self, ctx, cmd_name):
        if cmd_name == "create-mapped":
            cmd_name = "create"
        return super().get_command(ctx, cmd_name)


def globus_group(*args, **kwargs):
    disable_opts = kwargs.pop("disable_options", [])
    top_level_group = kwargs.pop("top_level", False)
    collection_group = kwargs.pop("collection_group", False)

    if top_level_group:
        kwargs["cls"] = TopLevelGroup

    if collection_group:
        kwargs["cls"] = CollectionCommandGroup

    def _inner_decorator(func):
        if "help" not in kwargs:
            kwargs["help"] = func.__doc__

        return _common_options(click.group(*args, **kwargs)(func), disable_opts)

    return _inner_decorator
