from typing import (
    ClassVar,
    List,
    Type,
    TYPE_CHECKING,
)

from .base import (
    Check,
    CheckInvoker,
    MissingPropertyCheck,
    ObjectIsNotNoneCheck,
    PropertyConstraintCheck,
)

if TYPE_CHECKING:
    from ..models.storage_gateway import StorageGateway  # noqa: F401


class StorageGatewayCheck(CheckInvoker["StorageGateway"]):
    checkers: ClassVar[List[Type[Check["StorageGateway"]]]] = []
    object_name = "storage_gateway"


@StorageGatewayCheck.checker
class SG01(ObjectIsNotNoneCheck["StorageGateway"]):
    ...


@StorageGatewayCheck.property_checker(property_name="display_name")
class SG02(MissingPropertyCheck["StorageGateway"]):
    ...


@StorageGatewayCheck.property_checker(property_name="display_name")
class SG03(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = "value contains all whitespace"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if value.display_name is not None:
            display_name = value.display_name.strip()
            if not display_name:
                ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="allowed_domains")
class SG04(MissingPropertyCheck["StorageGateway"]):
    ...


@StorageGatewayCheck.property_checker(property_name="allowed_domains")
class SG05(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = "must be set"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if value.allowed_domains is not None:
            allowed_domains = value.allowed_domains
            if len(allowed_domains) == 0:
                ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="allowed_domains")
class SG06(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = ">1 allowed_domains without identity mappings"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if value.allowed_domains is not None:
            allowed_domains = value.allowed_domains
            if len(allowed_domains) > 1 and not value.identity_mappings:
                ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="load_dsi_module")
class SG07(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = "load_dsi_module"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if value.load_dsi_module is not None and value.load_dsi_module != "file":
            ok = False
        return ok
