import json
import os
from typing import cast, Dict, List, Optional, Set, Tuple, TYPE_CHECKING

import click

from globus_sdk import GlobusError, IdentityMap

from ..models import MigrationPlan
from ..util import get_api
from .decorators import requires_migration_plan

if TYPE_CHECKING:
    from ..models import MyProxyConfiguration  # noqa: F401
    from ..typing.gcs_documents import IdentityMappingDict  # noqa: F401


@click.command("myproxy", help="Configure Authorization for MyProxy")
@click.option(
    "--subject",
    "-s",
    metavar="SUBJECT",
    help="MyProxy CA Subject. Optional if only one MyProxy CA is configured for "
    "your endpoint.",
)
@click.option(
    "--domain",
    "-d",
    multiple=True,
    required=True,
    metavar="DOMAIN",
    help="Domains to allow access to the endpoint. If "
    "specified, then only the specified domains are allowed access.",
)
@click.option(
    "--identity-mapping",
    "-i",
    multiple=True,
    metavar="JSON_OR_FILE",
    help="Either a JSON object or the name of a file that contains a json object. "
    "This object is parsed as an identity mapping document and added to the "
    "endpoint's policies. This is optional if there is only one domain being added, "
    "otherwise one must be defined to disambiguate how users are mapped.",
)
@click.option(
    "--dont-update-guest-collections",
    "-D",
    is_flag=True,
    help="If specified, don't update guest collection configurations to use "
    "this mapping.",
)
@requires_migration_plan
def identity_mapping_myproxy(
    subject: Optional[str],
    domain: Tuple[str],
    identity_mapping: Optional[str],
    dont_update_guest_collections: bool,
    migration_plan: MigrationPlan,
) -> None:
    if not migration_plan:
        raise click.ClickException("Unable to load migration plan")
    if not migration_plan.gsi_authz_configuration:
        raise click.ClickException("Migration plan has no GSI configuration")
    if not migration_plan.gsi_authz_configuration.myproxy_callouts:
        raise click.ClickException("Migration plan has no MyProxy configuration")
    if not migration_plan.storage_gateway:
        raise click.ClickException("Migration plan has no storage gateway")

    myproxy_configurations: Dict[int, "MyProxyConfiguration"] = {}

    # Load identity mapping if specified on the command-line
    identity_mapping_json: Optional[List["IdentityMappingDict"]] = None
    if identity_mapping:
        identity_mapping_json = []
        for idm in identity_mapping:
            if os.path.exists(idm):
                with open(idm, "r") as f:
                    identity_mapping_json.append(json.load(f))
            else:
                identity_mapping_json.append(json.loads(idm))

    # Extract MyProxy configurations from the authz config, and store them
    # along with their indices so we can check them for default values and then
    # update credentials which use these configurations.
    assert migration_plan.gsi_authz_configuration.callouts is not None
    for idx, callout in enumerate(migration_plan.gsi_authz_configuration.callouts):
        if callout.is_myproxy:
            myproxy_callout = cast("MyProxyConfiguration", callout)
            if not subject or subject == myproxy_callout.dn:
                myproxy_configurations[idx] = myproxy_callout

    # Determine subject to use if not specified on the command-line
    subjects: Set[str] = set()
    if not subject:
        for myproxy_callout in myproxy_configurations.values():
            if myproxy_callout.dn:
                subjects.add(myproxy_callout.dn)
        if len(subjects) > 1:
            raise click.ClickException(
                "Your GCSv4 configuration is configured to use more than one "
                "MyProxy CA. Please use the -s command-line option "
                "to select which one you will be adding the mapping for."
            )
        subject = subjects.pop()
    else:
        # Verify that the MyProxy subject was used by a configuration
        found_subject = False
        for myproxy_callout in myproxy_configurations.values():
            if myproxy_callout.dn == subject:
                found_subject = True
        if not found_subject:
            raise click.ClickException(
                f"Your endpoint has no MyProxy configuration for {subject}"
            )

    # Add the domains and identity mappings to the storage gateway
    migration_plan.storage_gateway.add_allowed_domains(
        list(domain), identity_mappings=identity_mapping_json
    )

    if migration_plan.guest_collections:
        # Update the credentials which match this MyProxy configuration to use
        # names if there is only one domain
        if dont_update_guest_collections:
            click.echo(
                "Not updating guest collection ownership. Use gcs-migration "
                "set-guest-collection-owner to update guest collections with "
                "appropriate ownership."
            )
            return
        elif len(domain) > 1:
            click.echo(
                "Not updating guest collection ownership because more than one "
                "domain is being allowed. Use gcs-migration "
                "set-guest-collection-owner to update guest collections with "
                "appropriate ownership."
            )
            return

        cred_domain = domain[0]

        # Create and populate an identity map to batch ID lookups
        identity_map = IdentityMap(get_api("auth.globus.org"))
        for gc in migration_plan.guest_collections:
            cred = gc.credential

            if cred and cred.gsi_authz_callout_index in myproxy_configurations.keys():
                assert cred.gsi_authz_callout_index is not None
                if cred.identity_username is None:
                    identity_username = f"{cred.username}@{cred_domain}"
                    identity_map.add(identity_username)
                myproxy_configurations[cred.gsi_authz_callout_index].add_domain(
                    cred_domain
                )

        # Resolve identities and update user credentials for guest collections.
        for gc in migration_plan.guest_collections:
            cred = gc.credential
            if cred and cred.gsi_authz_callout_index in myproxy_configurations.keys():
                if cred.identity_username is None:
                    identity_username = f"{cred.username}@{cred_domain}"
                    try:
                        identity = identity_map[identity_username]
                        cred.identity_username = identity_username
                        cred.identity_id = identity["id"]
                    except (GlobusError, KeyError):
                        click.echo(
                            f"Unable to resolve {identity_username} to a valid "
                            f"Globus identity, not updating configuration for {gc.id}"
                        )
    migration_plan.dump_storage_gateway()
    migration_plan.dump_storage_gateway()
    migration_plan.dump_guest_collections()
    migration_plan.dump_gsi_authz_configuration()
