import click

import globus_sdk.exc


from gcs_migration.config import (
    AUTH_CLIENT_DATA_CONFIG_KEY,
    internal_auth_client,
    token_storage_adapter,
)

from ..util import do_link_auth_flow


@click.command(
    "login", short_help="Get credentials to run the GCS Migration tools",
)
@click.option(
    "--force",
    is_flag=True,
    help="Do a fresh auth flow, ignoring any existing credentials",
)
def login(force: bool) -> None:
    if not force and check_logged_in():
        return

    do_link_auth_flow(session_params={"session_required_single_domain": "globusid.org"})


def check_logged_in() -> bool:
    """
    perform a series of checks to get a complete/working "logged in" state
    if any check fails, return False
    """
    adapter = token_storage_adapter()
    # first, get the GCS client info (so we can lookup tokens)
    auth_client_data = adapter.read_config(AUTH_CLIENT_DATA_CONFIG_KEY)
    if not auth_client_data:
        return False

    # try to get the tokens from config
    services = ["auth.globus.org", "transfer.api.globus.org"]
    for service in services:
        token_data = adapter.get_token_data(service)
        if not token_data:
            return False

        # pull out the refresh token from there (at this point, the storage adapter
        # should be giving us a complete/working response unless someone messed
        # with the DB, but let's play it safe and check for null
        refresh_token = token_data.get("refresh_token")
        if refresh_token is None:
            return False

        # get or create the instance client
        auth_client = internal_auth_client()

        # check that token and client are valid
        try:
            res = auth_client.oauth2_validate_token(refresh_token)
            if not res["active"]:
                return False

        # if the instance client is invalid, an AuthAPIError will be raised
        # we then force a new client to be created before continuing
        except globus_sdk.AuthAPIError:
            return False

    return True
