import click

from .create import create
from gcs_migration.models import MigrationPlan
from gcs_migration.exceptions import UnsupportedUpdate


@click.command(
    "update",
    short_help="Merge any endpoint changes to the migration plan",
    help="""\
Check the endpoint for configuration changes since the migration plan was
created, and then incorporate these changes into the migration plan.
""",
)
def update() -> None:
    old_migration_plan = MigrationPlan.load()

    context = click.get_current_context()
    new_migration_plan = context.invoke(create, no_write=True)

    try:
        # Ignore the gridftp_config and gcs_config for now. The
        # actionable parts are incorporated in the new migration plan's storage
        # gateway and/or collection models
        assert old_migration_plan.gsi_authz_configuration is not None
        assert new_migration_plan.gsi_authz_configuration is not None
        if (
            old_migration_plan.gsi_authz_configuration
            != new_migration_plan.gsi_authz_configuration
        ):
            raise UnsupportedUpdate("Unsupported update to gsi-authz.conf")

        assert old_migration_plan.storage_gateway is not None
        assert new_migration_plan.storage_gateway is not None
        old_migration_plan.storage_gateway.update(new_migration_plan.storage_gateway)

        assert old_migration_plan.mapped_collection is not None
        assert new_migration_plan.mapped_collection is not None
        old_migration_plan.mapped_collection.update(
            new_migration_plan.mapped_collection
        )
        assert old_migration_plan.mapped_collection is not None
        assert new_migration_plan.mapped_collection is not None
        old_migration_plan.mapped_collection.update(
            new_migration_plan.mapped_collection
        )

        if old_migration_plan.environment != new_migration_plan.environment:
            raise UnsupportedUpdate("Unsupported environment change")

        new_guest_collections = getattr(new_migration_plan, "guest_collections", [])
        old_guest_collections = getattr(old_migration_plan, "guest_collections", [])

        old_guest_collections = {g.id: g for g in old_guest_collections}
        new_guest_collections = {g.id: g for g in new_guest_collections}

        if old_migration_plan.guest_collections:
            for gc in old_guest_collections.values():
                if gc.id not in new_guest_collections:
                    click.echo(f"Removing {gc.id}")
                    old_migration_plan.remove_guest_collection(gc)
                else:
                    click.echo(f"Updating {gc.id}")
                    gc.update(new_guest_collections[gc.id])

        for gc in new_guest_collections.values():
            if gc.id not in old_guest_collections:
                click.echo(f"Adding {gc}")
                old_migration_plan.add_guest_collection(gc)

        if old_migration_plan.guest_collections:
            for gc in old_migration_plan.guest_collections:
                if old_migration_plan.mapped_collection.force_verify:
                    gc.force_verify = True
                    gc.disable_verify = False
                if old_migration_plan.mapped_collection.force_encryption:
                    gc.force_encryption = True
    except UnsupportedUpdate as e:
        raise click.ClickException(str(e))

    old_migration_plan.dump()
