from typing import List, Optional, TYPE_CHECKING
from typing_extensions import TypedDict

from ..util import yaml, yaml_object

if TYPE_CHECKING:
    from ..typing.transfer_documents import EndpointDocumentDict  # noqa: F401
    from ..typing.gcs_documents import _MetadataDict  # noqa: F401


class EndpointMetadataDict(TypedDict):
    id: Optional[str]
    display_name: Optional[str]
    public: Optional[bool]
    description: Optional[str]
    info_link: Optional[str]
    contact_info: Optional[str]
    contact_email: Optional[str]
    organization: Optional[str]
    department: Optional[str]
    keywords: Optional[List[str]]


@yaml_object(yaml)
class EndpointMetadata:
    def __init__(self) -> None:
        self.id: Optional[str] = None
        self.display_name: Optional[str] = None
        self.public: Optional[bool] = None
        self.description: Optional[str] = None
        self.info_link: Optional[str] = None
        self.contact_info: Optional[str] = None
        self.contact_email: Optional[str] = None
        self.organization: Optional[str] = None
        self.department: Optional[str] = None
        self.keywords: Optional[List[str]] = None
        self._transfer_endpoint = None

    @property
    def _transfer_endpoint(self) -> Optional["EndpointDocumentDict"]:
        return self.__transfer_endpoint

    @_transfer_endpoint.setter
    def _transfer_endpoint(self, value: Optional["EndpointDocumentDict"]) -> None:
        self.__transfer_endpoint = value
        if value:
            if self.id is None:
                self.id = value["id"]
            if self.display_name is None:
                self.display_name = value["display_name"] or value["canonical_name"]
            if self.public is None:
                self.public = value["public"]
            if self.description is None:
                self.description = value["description"]
            if self.info_link is None:
                self.info_link = value["info_link"]
            if self.contact_info is None:
                self.contact_info = value["contact_info"]
            if self.contact_email is None:
                self.contact_email = value["contact_email"]
            if self.organization is None:
                self.organization = value["organization"]
            if self.department is None:
                self.department = value["department"]
            if self.keywords is None:
                kw = value["keywords"]
                if kw:
                    self.keywords = [s.strip() for s in kw.split(",")]
                else:
                    self.keywords = None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, EndpointMetadata):
            return NotImplemented

        return self is other or self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> EndpointMetadataDict:
        return {
            "id": self.id,
            "display_name": self.display_name,
            "public": self.public,
            "description": self.description,
            "info_link": self.info_link,
            "contact_info": self.contact_info,
            "contact_email": self.contact_email,
            "organization": self.organization,
            "department": self.department,
            "keywords": self.keywords,
        }

    def __setstate__(self, state: EndpointMetadataDict) -> None:
        self.id = state["id"]
        self.display_name = state["display_name"]
        self.public = state["public"]
        self.description = state["description"]
        self.info_link = state["info_link"]
        self.contact_info = state["contact_info"]
        self.contact_email = state["contact_email"]
        self.organization = state["organization"]
        self.department = state["department"]
        self.keywords = state["keywords"]

    def to_json(self) -> "_MetadataDict":
        metadata: "_MetadataDict" = {}

        if self.display_name:
            metadata["display_name"] = self.display_name
        if self.description:
            metadata["description"] = self.description
        if self.info_link:
            metadata["info_link"] = self.info_link
        if self.contact_info:
            metadata["contact_info"] = self.contact_info
        if self.contact_email:
            metadata["contact_email"] = self.contact_email
        if self.organization:
            metadata["organization"] = self.organization
        if self.department:
            metadata["department"] = self.department
        if self.keywords:
            metadata["keywords"] = self.keywords
        return metadata

    def update(self, other: "EndpointMetadata") -> None:
        for attr in [
            "id",
            "display_name",
            "public",
            "description",
            "info_link",
            "contact_info",
            "contact_email",
            "organization",
            "department",
            "keywords",
        ]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if old_value != new_value:
                setattr(self, attr, new_value)
