import os
import pathlib
from typing import Iterable, List, Optional, Union
from typing_extensions import TypedDict

from ..parsers import GridFTPOptionParser, parse_restrict_paths
from ..typing.gcs_documents import (
    IdentityMappingDict,
    PathRestrictionsDict,
    PosixStorageGatewayDict,
    PosixStorageGatewayPoliciesDict,
)
from ..util import yaml, yaml_object
from ..exceptions import UnsupportedUpdate


class StorageGatewayDict(TypedDict):
    display_name: Optional[str]
    users_allow: Optional[List[str]]
    users_deny: Optional[List[str]]
    groups_allow: Optional[List[str]]
    groups_deny: Optional[List[str]]
    load_dsi_module: Optional[str]
    allowed_domains: Optional[List[str]]
    identity_mappings: Optional[List[IdentityMappingDict]]
    restrict_paths: Optional[PathRestrictionsDict]


@yaml_object(yaml)
class StorageGateway:
    def __init__(self, data: Optional[StorageGatewayDict] = None) -> None:
        self.display_name: Optional[str] = None
        self.users_allow: Optional[List[str]] = None
        self.users_deny: Optional[List[str]] = None
        self.groups_allow: Optional[List[str]] = None
        self.groups_deny: Optional[List[str]] = None
        self.load_dsi_module: Optional[str] = None
        self.allowed_domains: Optional[List[str]] = None
        self.identity_mappings: Optional[List[IdentityMappingDict]] = None
        self.restrict_paths: Optional[PathRestrictionsDict] = None
        self.gridftp_config: Optional[GridFTPOptionParser] = None

        if data:
            self.__setstate__(data)

    @property
    def gridftp_config(self) -> Optional[GridFTPOptionParser]:
        return self._gridftp_config

    @gridftp_config.setter
    def gridftp_config(self, value: Optional[GridFTPOptionParser]) -> None:
        self._gridftp_config = value
        if value:
            if self.restrict_paths is None:
                rp = value.get_string_option("restrict_paths")
                if rp:
                    self.restrict_paths = parse_restrict_paths(rp)
            if self.load_dsi_module is None:
                self.load_dsi_module = value.get_string_option("load_dsi_module")

    @property
    def restrict_paths(self) -> Optional[PathRestrictionsDict]:
        return self._restrict_paths

    @restrict_paths.setter
    def restrict_paths(self, value: Optional[PathRestrictionsDict]) -> None:
        self._restrict_paths = value

    @property
    def allowed_domains(self) -> Optional[List[str]]:
        return self._allowed_domains

    @allowed_domains.setter
    def allowed_domains(self, value: Optional[List[str]]) -> None:
        self._allowed_domains = value

    def dump(self, root: Union[str, os.PathLike]) -> None:
        path = pathlib.Path(root) / "storage_gateway.yaml"
        with open(path, "w") as f:
            yaml.dump(self, f)

    @staticmethod
    def load(root: Union[str, os.PathLike]) -> Optional["StorageGateway"]:
        path = pathlib.Path(root) / "storage_gateway.yaml"
        if not os.path.exists(path):
            return None

        with open(path, "r") as f:
            obj = yaml.load(f)
            assert isinstance(obj, StorageGateway)

        return obj

    def add_allowed_domains(
        self,
        domains: Iterable[str],
        identity_mappings: Optional[List[IdentityMappingDict]] = None,
    ) -> None:
        if self.allowed_domains is None:
            self.allowed_domains = []
        if self.identity_mappings is None:
            self.identity_mappings = []

        for domain in domains:
            if domain not in self.allowed_domains:
                self.allowed_domains.append(domain)

        if identity_mappings is None:
            domains_list = "|".join([d.replace(".", "\\.") for d in domains])
            identity_mappings = [
                {
                    "DATA_TYPE": "expression_identity_mapping#1.0.0",
                    "mappings": [
                        {
                            "source": "{username}",
                            "match": f"(.*)@({domains_list})",
                            "output": "{0}",
                        }
                    ],
                }
            ]
        for identity_mapping in identity_mappings:
            if identity_mapping not in self.identity_mappings:
                self.identity_mappings.append(identity_mapping)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StorageGateway):
            return NotImplemented

        return self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> StorageGatewayDict:
        return {
            "display_name": self.display_name,
            "users_allow": self.users_allow,
            "users_deny": self.users_deny,
            "groups_allow": self.groups_allow,
            "groups_deny": self.groups_deny,
            "load_dsi_module": self.load_dsi_module,
            "allowed_domains": self.allowed_domains,
            "identity_mappings": self.identity_mappings,
            "restrict_paths": self.restrict_paths,
        }

    def __setstate__(self, state: StorageGatewayDict) -> None:
        self.display_name = state["display_name"]
        self.users_allow = state["users_allow"]
        self.users_deny = state["users_deny"]
        self.groups_allow = state["groups_allow"]
        self.groups_deny = state["groups_deny"]
        self.load_dsi_module = state["load_dsi_module"]
        self.allowed_domains = state["allowed_domains"]
        self.identity_mappings = state["identity_mappings"]
        self.restrict_paths = state["restrict_paths"]

    def to_json(self) -> PosixStorageGatewayDict:
        policies: PosixStorageGatewayPoliciesDict = {
            "DATA_TYPE": "posix_storage_policies#1.0.0"
        }

        storage_gateway: PosixStorageGatewayDict = {
            "DATA_TYPE": "storage_gateway#1.1.0",
            "connector_id": "145812c8-decc-41f1-83cf-bb2a85a2a70b",
            "high_assurance": False,
            "policies": policies,
        }

        if self.display_name:
            storage_gateway["display_name"] = self.display_name

        if self.allowed_domains:
            storage_gateway["allowed_domains"] = self.allowed_domains

        if self.identity_mappings:
            storage_gateway["identity_mappings"] = self.identity_mappings

        if self.users_allow:
            storage_gateway["users_allow"] = self.users_allow

        if self.users_deny:
            storage_gateway["users_deny"] = self.users_deny

        if self.restrict_paths:
            storage_gateway["restrict_paths"] = self.restrict_paths

        if self.groups_allow:
            policies["groups_allow"] = self.groups_allow

        if self.groups_deny:
            policies["groups_deny"] = self.groups_deny
        return storage_gateway

    def update(self, other: "StorageGateway") -> None:
        for attr in [
            "display_name",
            "users_allow",
            "users_deny",
            "groups_allow",
            "groups_deny",
            "restrict_paths",
        ]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if old_value != new_value:
                setattr(self, attr, new_value)

        unsupported_updates = []
        for attr in ["load_dsi_module", "identity_mappings", "allowed_domains"]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if new_value and old_value != new_value:
                unsupported_updates.append(
                    f"Unable to update {attr} from {old_value} to {new_value}"
                )
        if unsupported_updates:
            raise UnsupportedUpdate("\n".join(unsupported_updates))

    def __str__(self) -> str:
        if self.display_name:
            return f"<StorageGateway {self.display_name}>"
        else:
            return "<StorageGateway>"
