from typing import Optional
import uuid

import configparser
import os


class GCSConfig:
    def __init__(self, path: Optional[str] = None, endpoint_id: Optional[str] = None):
        if path is None:
            path = "/etc/globus-connect-server.conf"

        defaults = {
            "GLOBUS_USER": "",
            "GLOBUS_PASSWORD": "",
            "SHORT_HOSTNAME": "",
            "HOSTNAME": "",
        }
        config = configparser.ConfigParser(defaults=defaults)
        config.read(path)
        self.config = config

        self.user: Optional[str] = config.get("Globus", "User", fallback=None)
        self.password: Optional[str] = config.get("Globus", "Password", fallback=None)
        self.instance: str = config.get("Globus", "Instance", fallback="")
        if not self.instance:
            environment = os.environ.get("GLOBUS_SDK_ENVIRONMENT")
            if environment:
                self.instance = environment
            else:
                self.instance = "default"

        self.endpoint_id = endpoint_id

    @property
    def endpoint_id(self) -> Optional[str]:
        return self._endpoint_id

    @endpoint_id.setter
    def endpoint_id(self, value: Optional[str]) -> None:
        if value is None:
            uuid_path = "/var/lib/globus-connect-server/gcsv4-endpoint-id.txt"
            v4_endpoint_uuid = None
            try:
                if os.path.exists(uuid_path):
                    with open(uuid_path, "r") as f:
                        v4_endpoint_uuid = str(uuid.UUID(f.read().strip()))
            except Exception:
                pass
            self._endpoint_id = v4_endpoint_uuid
        else:
            self._endpoint_id = value

    def get(
        self, section: str, field: str, fallback: Optional[str] = None
    ) -> Optional[str]:
        return self.config.get(section, field, fallback=fallback)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GCSConfig):
            return NotImplemented
        return self is other or (
            self.user == other.user and self.instance == other.instance
        )
