import ctypes
import ctypes.util
from multiprocessing import Process, Queue
import os
from typing import Dict, List, Optional, Union

from .gridftp_options import options


class GridFTPOptionParser:
    @classmethod
    def _parse_options(
        cls,
        args: List[Union[str, os.PathLike]],
        response_queue: "Queue[Union[Dict, Exception]]",
    ) -> None:
        globus_common = ctypes.CDLL(ctypes.util.find_library("globus_common"))
        globus_gridftp_server = ctypes.CDLL(
            ctypes.util.find_library("globus_gridftp_server")
        )

        GLOBUS_COMMON_MODULE = globus_common.globus_i_common_module
        GLOBUS_GRIDFTP_SERVER_MODULE = globus_gridftp_server.globus_i_gfs_module
        globus_module_activate = globus_common.globus_module_activate
        globus_module_activate.argtypes = [ctypes.c_void_p]
        globus_module_activate.restype = ctypes.c_int32

        globus_gfs_config_get_bool = globus_gridftp_server.globus_gfs_config_get_bool
        globus_gfs_config_get_bool.argtypes = [ctypes.c_char_p]
        globus_gfs_config_get_bool.restype = ctypes.c_int

        globus_gfs_config_get_int = globus_gridftp_server.globus_gfs_config_get_int
        globus_gfs_config_get_int.argtypes = [ctypes.c_char_p]
        globus_gfs_config_get_int.restype = ctypes.c_int

        globus_i_gfs_config_init_envs = (
            globus_gridftp_server.globus_i_gfs_config_init_envs
        )
        globus_i_gfs_config_init_envs.argtypes = [
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_char_p),
        ]
        globus_i_gfs_config_init_envs.restype = ctypes.c_int

        globus_i_gfs_config_init = globus_gridftp_server.globus_i_gfs_config_init
        globus_i_gfs_config_init.argtypes = [
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_char_p),
            ctypes.c_int,
        ]
        globus_i_gfs_config_init.restype = ctypes.c_int

        globus_gfs_config_get_string = globus_gridftp_server.globus_gfs_config_get_list
        globus_gfs_config_get_string.argtypes = [ctypes.c_char_p]
        globus_gfs_config_get_string.restype = ctypes.c_char_p

        globus_libc_getenv = globus_common.globus_libc_getenv
        globus_libc_getenv.argtypes = [ctypes.c_char_p]
        globus_libc_getenv.restype = ctypes.c_char_p

        argc = len(args) + 1
        argv = (ctypes.c_char_p * argc)()
        argv[0] = ctypes.c_char_p("X".encode("ascii"))
        for idx, a in enumerate(args):
            if isinstance(a, os.PathLike):
                path = a.__fspath__()
                if isinstance(path, str):
                    argv[idx + 1] = ctypes.c_char_p(path.encode("ascii"))
                else:
                    argv[idx + 1] = ctypes.c_char_p(path)
            else:
                argv[idx + 1] = ctypes.c_char_p(a.encode("ascii"))

        rc = globus_i_gfs_config_init_envs(ctypes.c_int(argc), argv)
        if rc != 0:
            response_queue.put(ValueError("Init envs error"))
            return
        rc = globus_module_activate(GLOBUS_COMMON_MODULE)
        if rc != 0:
            response_queue.put(ValueError("Activation error"))
            return

        rc = globus_module_activate(GLOBUS_GRIDFTP_SERVER_MODULE)
        if rc != 0:
            response_queue.put(ValueError("Activation error"))
            return

        rc = globus_i_gfs_config_init(ctypes.c_int(argc), argv, ctypes.c_int(0))
        if rc != 0:
            response_queue.put(ValueError("parsing error"))
            return

        response = {}
        for option_spec in options.values():
            if option_spec.name.startswith("$"):
                s = globus_libc_getenv(
                    ctypes.c_char_p(option_spec.name[1:].encode("ascii"))
                )
                if s:
                    response[option_spec.name] = s.decode("ascii")
                else:
                    response[option_spec.name] = None
            elif option_spec:
                if option_spec.type == "string":
                    s = globus_gfs_config_get_string(option_spec.name.encode("ascii"))
                    if s:
                        response[option_spec.name] = s.decode("ascii")
                    else:
                        response[option_spec.name] = None
                elif option_spec.type == "bool":
                    response[option_spec.name] = bool(
                        globus_gfs_config_get_bool(option_spec.name.encode("ascii"))
                    )
                elif option_spec.type == "int":
                    response[option_spec.name] = globus_gfs_config_get_int(
                        option_spec.name.encode("ascii")
                    )

        response_queue.put(response)

    def __init__(
        self,
        args: Optional[List[Union[str, os.PathLike]]] = None,
        config_file: Union[str, os.PathLike] = None,
        config_dir: Union[str, os.PathLike] = None,
    ) -> None:
        if args is None:
            args = []

        if config_file is not None:
            if not os.path.exists(config_file):
                raise FileNotFoundError(config_file)
            args.extend(["-c", config_file])

        if config_dir is not None:
            if not os.path.exists(config_dir):
                raise FileNotFoundError(config_dir)
            args.extend(["-C", config_dir])

        response_queue: "Queue[Union[Dict[str, Union[int, str, None]], Exception]]" = Queue()

        self.args = args
        if args:
            p = Process(target=self._parse_options, args=(args, response_queue,),)
            p.start()
            res = response_queue.get()
            p.join()
            if isinstance(res, Exception):
                raise res
            self.options = res
        else:
            self.options = {val.name: val.default for val in options.values()}

    def get_option(self, option_name: str) -> Union[str, bool, int, None]:
        return self.options[option_name]

    def get_bool_option(self, option_name: str) -> Optional[bool]:
        option_spec = options.get(option_name)
        if not option_spec:
            raise KeyError

        if option_spec.type != "bool":
            raise TypeError

        value = self.get_option(option_name)
        assert value is None or isinstance(value, bool)

        return value

    def get_int_option(self, option_name: str) -> Optional[int]:
        option_spec = options.get(option_name)
        if not option_spec:
            raise KeyError

        if option_spec.type != "int":
            raise TypeError

        value = self.get_option(option_name)
        assert value is None or isinstance(value, int)
        return value

    def get_string_option(self, option_name: str) -> Optional[str]:
        option_spec = options.get(option_name)
        if not option_spec:
            raise KeyError

        if option_spec.type != "string":
            raise TypeError

        value = self.get_option(option_name)
        assert value is None or isinstance(value, str)
        return value

    def get_environment(self) -> Dict[str, str]:
        result: Dict[str, str] = {}

        for opt in options.values():
            if opt.name.startswith("$"):
                value = self.get_string_option(opt.name)
                if value:
                    result[opt.name[1:]] = str(value)

        return result
