import ctypes
import ctypes.util
from multiprocessing import Process, Queue
import os
import tempfile
from typing import Dict, Optional, Tuple, TYPE_CHECKING

from cryptography import x509

if TYPE_CHECKING:
    from gcs_migration.models.gsi_authz_configuration import GsiAuthzConfiguration


class AuthorizeSharing:
    def __init__(self, gsi_authz_configuration: "GsiAuthzConfiguration"):
        self.gsi_authz_configuration = gsi_authz_configuration

    def _map_and_authorize_sharing(
        self,
        line: str,
        sharing_cert: str,
        environment: Dict[str, str],
        responder: Queue,
    ) -> None:
        globus_common = ctypes.CDLL(ctypes.util.find_library("globus_common"))
        globus_gss_assist = ctypes.CDLL(ctypes.util.find_library("globus_gss_assist"))

        GLOBUS_GSS_ASSIST_MODULE = globus_gss_assist.globus_i_gsi_gss_assist_module
        GLOBUS_COMMON_MODULE = globus_gss_assist.globus_i_common_module

        globus_module_activate = globus_common.globus_module_activate
        globus_module_activate.argtypes = [ctypes.c_void_p]
        globus_module_activate.restype = ctypes.c_int

        globus_module_deactivate = globus_common.globus_module_deactivate
        globus_module_deactivate.argtypes = [ctypes.c_void_p]
        globus_module_deactivate.restype = ctypes.c_int

        globus_libc_setenv = globus_common.globus_libc_setenv
        globus_libc_setenv.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int]
        globus_libc_setenv.restype = ctypes.c_int

        globus_libc_unsetenv = globus_common.globus_libc_unsetenv
        globus_libc_unsetenv.argtypes = [ctypes.c_char_p]
        globus_libc_unsetenv.restype = None

        _globus_gss_assist_map_and_authorize_sharing = (
            globus_gss_assist.globus_gss_assist_map_and_authorize_sharing
        )
        _globus_gss_assist_map_and_authorize_sharing.argtypes = [
            ctypes.c_char_p,
            ctypes.c_void_p,
            ctypes.c_void_p,
            ctypes.c_char_p,
            ctypes.c_uint,
        ]
        _globus_gss_assist_map_and_authorize_sharing.restype = ctypes.c_int32

        GSI_AUTHZ_CONF = ctypes.c_char_p("GSI_AUTHZ_CONF".encode("ascii"))
        GRIDMAP = ctypes.c_char_p("GRIDMAP".encode("ascii"))

        globus_module_activate(GLOBUS_COMMON_MODULE)
        globus_module_activate(GLOBUS_GSS_ASSIST_MODULE)
        name = None
        for k, v in environment.items():
            globus_libc_setenv(
                ctypes.c_char_p(k.encode("ascii")),
                ctypes.c_char_p(v.encode("ascii")),
                ctypes.c_int(1),
            )
        if line:
            (fd, name) = tempfile.mkstemp()
            os.write(fd, f"{line}\n".encode("ascii"))
            os.close(fd)

            globus_libc_setenv(
                GSI_AUTHZ_CONF, ctypes.c_char_p(name.encode("ascii")), ctypes.c_int(1),
            )
        else:
            globus_libc_unsetenv(GSI_AUTHZ_CONF)

        if self.gsi_authz_configuration.gridmap:
            globus_libc_setenv(
                GRIDMAP,
                ctypes.c_char_p(
                    str(self.gsi_authz_configuration.gridmap).encode("ascii")
                ),
                ctypes.c_int(1),
            )
        else:
            globus_libc_unsetenv(GRIDMAP)

        sharing_cert_cstr = ctypes.c_char_p(sharing_cert.encode("ascii"))
        authz_user_cstr = (ctypes.c_char * 64)()
        authz_user_len = ctypes.c_uint(64)
        res = _globus_gss_assist_map_and_authorize_sharing(
            sharing_cert_cstr, None, None, authz_user_cstr, authz_user_len
        )
        if name:
            os.remove(name)

        if res == 0:
            responder.put(authz_user_cstr.value.decode("ascii"))
        else:
            responder.put(None)
        globus_module_deactivate(GLOBUS_GSS_ASSIST_MODULE)
        globus_module_deactivate(GLOBUS_COMMON_MODULE)

    def map_and_authorize_sharing(
        self, sharing_cert: str
    ) -> Tuple[Optional[str], str, Optional[int]]:
        # Run in subprocess to deal with caching in the callout library making
        # it difficult to run against individual lines of the authz config file
        # to figure out which mapping occurred.
        cert = x509.load_pem_x509_certificate(sharing_cert.encode("ascii"))
        subject = "/" + "/".join([rdn.rfc4514_string() for rdn in cert.subject.rdns])
        if (
            self.gsi_authz_configuration
            and self.gsi_authz_configuration.callout_dict_list
        ):
            for idx, conf in enumerate(self.gsi_authz_configuration.callout_dict_list):
                q: Queue = Queue()
                p = Process(
                    target=self._map_and_authorize_sharing,
                    args=(
                        conf["raw"] if conf else None,
                        sharing_cert,
                        conf["environment"]
                        if conf
                        else self.gsi_authz_configuration.environment,
                        q,
                    ),
                )
                p.start()
                res = q.get()
                p.join()
                if res:
                    return (res, subject, idx)
        return (None, subject, None)
