import os
import sys
from typing import Optional, Tuple

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))  # noqa: E402

import globus.manager.version  # noqa: E402


class GcsVersion:
    STATUS_VALUES = ["alpha", "beta", "rc", "release"]

    def __init__(self):
        fingerprint = globus.manager.version.__version_fingerprint__
        self.major = fingerprint["major"]
        self.minor = fingerprint["minor"]
        self.patch = fingerprint["patch"]
        self._status, self._status_version = self._parse_status(fingerprint["status"])
        self._api_version = globus.manager.version.__api_version__

    @property
    def status(self) -> str:
        return self._status

    @status.setter
    def status(self, value: str):
        status_index = self.STATUS_VALUES.index(self._status)
        new_status_index = self.STATUS_VALUES.index(value)
        if new_status_index > status_index:
            # New status value in the same release version
            self._status = value
            if value == "release":
                self._status_version = None
            else:
                self._status_version = 1
        elif new_status_index == status_index and value != "release":
            # Same non-release status, update status version
            self._status_version += 1
        else:
            # Either same release status, or a jump backward in status index
            self.patch += 1
            self._status = value
            if value == "release":
                self._status_version = None
            else:
                self._status_version = 1

    @property
    def api_version(self):
        return self._api_version

    @api_version.setter
    def api_version(self, value: str):
        old_major, old_minor, old_patch = [int(v) for v in self._api_version.split(".")]
        major, minor, patch = [int(v) for v in value.split(".")]

        if major == old_major:
            if minor == old_minor:
                if patch == old_patch:
                    return
                elif patch > old_patch:
                    self._api_version = value
                    return
            elif minor > old_minor:
                self._api_version = value
                return
        elif major > old_major:
            self._api_version = value
            return
        raise ValueError(
            "API Version {} does not update {}".format(self._api_version, value)
        )

    @staticmethod
    def previous_version():
        previous = GcsVersion()
        previous._status = "release"
        previous._status_version = None
        previous.patch -= 1
        return previous

    @staticmethod
    def _parse_status(status_string: str) -> Optional[Tuple[str, int]]:
        if status_string.startswith("a"):
            status = "alpha"
            status_version = int(status_string.split("a")[1])
        elif status_string.startswith("b"):
            status = "beta"
            status_version = int(status_string.split("b")[1])
        elif status_string.startswith("rc"):
            status = "rc"
            status_version = int(status_string.split("rc")[1])
        elif status_string.startswith("release"):
            status = "release"
            status_version = None
        else:
            return None

        return (status, status_version)

    def __render_status(self):
        status_format = {
            "alpha": "-a{status_version}",
            "beta": "-b{status_version}",
            "rc": "-rc{status_version}",
            "release": "",
        }
        return status_format[self._status].format(status_version=self._status_version)

    def version(self):
        return "{}.{}.{}{}".format(
            self.major, self.minor, self.patch, self.__render_status()
        )

    def __str__(self):
        version = "{}.{}.{}{}".format(
            self.major, self.minor, self.patch, self.__render_status()
        )
        return """\
# DO NOT EDIT: Generated by packaging/update-package-info
# single source of truth for package version,
# see https://packaging.python.org/en/latest/single_source_version/
__version__ = "{version}"
__api_version__ = "{api_version}"

__version_split__ = __version__.split(".", 2)

__version_fingerprint__ = {{
    "major": int(__version_split__[0]),
    "minor": int(__version_split__[1]),
    "patch": int(__version_split__[2].split("-", 1)[0]),
    "status": (
        __version_split__[2].split("-", 1)[1]
        if "-" in __version_split__[2]
        else "release"
    ),
}}""".format(
            version=version, api_version=self.api_version
        )
