import time
from typing import Optional

from flask import current_app, g
from cachelib import SimpleCache

import globus.manager.api.errors as errors

import globus_sdk


cache = SimpleCache()


def _is_mfa(authentication):
    """
    Local helper function that determins if session authentication was
    performed using multi-factor authentication.
    """
    if "amr" not in authentication:
        return False
    if not isinstance(authentication["amr"], list):
        return False
    return "mfa" in authentication["amr"]


class AuthenticationState:
    """
    Authentication state containing information about the identities,
    sessions, and authentications associated with a token.  It can also
    be used to manage similar data coming from identity assertions
    from transfer or mod-globus.
    """

    def __init__(
        self,
        token=None,
        identity_set=None,
        identity_set_data=None,
        session_info=None,
        last_authentications=None,
    ):
        """
        Initialize a new AuthenticationState object. If the token parameter
        is provided it will be stored in the token property and used with
        Globus Auth to determine the identity_set, identity_set_data,
        session_authentications, and non_session_authentications. Otherwise,
        those values are passed in via the other parameters.

        :param token:
            An optional :py:code`str` containing access token used to authorize
            access to this service.
        :param identity_set:
            An optional list of :py:code`str` values, each an identity id
            linked to the effective identity calling this service.
        :param identity_set_data:
            A list of dicts containing the properties of a Globus Auth
            Identity.
        :param session_info:
            A dict containing the session_info property from a token
            introspection call.
        :param last_authentications:
            A dict mapping identity ID values to the ISO 8601 formatted
            timestamp when the user last authenticated via the IdP to Globus
            Auth in any application.
        """
        self.token = token
        self.is_authenticated = False
        if identity_set is None:
            identity_set = []
        self.identity_set = identity_set

        self._identity_set_data = identity_set_data

        if session_info:
            authentications = session_info.get("authentications", {})
            self._session_authentications = {
                k: v["auth_time"] for k, v in authentications.items()
            }
            self._multi_factor_auth = {
                k: _is_mfa(v) for k, v in authentications.items()
            }
        else:
            self._session_authentications = None
            self._multi_factor_auth = None

        self._non_session_authentications = {
            k: int(v) for k, v in (last_authentications or {}).items()
        }

        if self.token is not None:
            self.__introspect_token()
        elif self._identity_set_data and not self.identity_set:
            self.identity_set = [i["id"] for i in self._identity_set_data]

    def __introspect_token(self):
        assert self.token is not None

        introspect_cache_key = "introspect_{}".format(self.token)
        self.identity_set = None
        self._identity_set_data = None
        self._session_authentications = None
        self._non_session_authentications = None
        self._multi_factor_auth = None

        current_app.logger.debug(
            'Authenticating with token ending in "...{}"'.format(self.token[-5:])
        )

        token_info = cache.get(introspect_cache_key)
        if token_info is None:
            # TODO: Only check session_info if subscription has HA enabled
            token_info = current_app.auth_client.oauth2_token_introspect(
                self.token, include="identity_set,session_info"
            )

            cache.set(introspect_cache_key, token_info.data, timeout=60)

        if token_info.get("active"):
            current_app.logger.debug("Authentication success.")
            self.is_authenticated = True
            self.identity_set = token_info.get("identity_set", [])
            session_info = token_info.get("session_info", {})
            authentications = session_info.get("authentications", {})

            self._session_authentications = {
                k: v["auth_time"] for k, v in authentications.items()
            }
            self._multi_factor_auth = {
                k: _is_mfa(v) for k, v in authentications.items()
            }
        else:
            current_app.logger.debug("Authentication failed -- invalid token.")

    @property
    def identity_set_data(self):
        """
        A list of Globus Auth identity resources
        """
        if self.identity_set is None:
            return []

        if self._identity_set_data is None:
            idset_key = "identity_set_data_{}".format(
                "_".join(sorted(self.identity_set))
            )

            identity_set_data = cache.get(idset_key)

            if not identity_set_data:
                if self.identity_set:
                    r = current_app.auth_client.get_identities(ids=self.identity_set)
                    identity_set_data = r.get("identities")
                else:
                    identity_set_data = []

                cache.set(idset_key, identity_set_data, timeout=60)
            self._identity_set_data = identity_set_data

        return self._identity_set_data

    @property
    def session_authentications(self):
        """
        Dictionary mapping Globus Auth Identity IDs to an integer
        timestamp. The timestamp represents the last time the user
        authenticated with the identity in the session where the token was
        generated.
        """
        return self._session_authentications or {}

    @property
    def non_session_authentications(self):
        """
        Dictionary mapping Globus Auth Identity IDs to an integer
        timestamp. The timestamp represents the last time the user
        authenticated the identity via the identity provider to Globus Auth
        in any application.
        """
        if self.token and not self._non_session_authentications:
            userinfo_cache_key = "userinfo_{}".format(self.token)
            userinfo = cache.get(userinfo_cache_key)

            if not userinfo:
                token = self.dependent_token("auth.globus.org")
                ac = globus_sdk.AuthClient(
                    authorizer=globus_sdk.AccessTokenAuthorizer(token),
                    environment=current_app.environment,
                )
                userinfo = ac.oauth2_userinfo().data
                cache.set(userinfo_cache_key, userinfo, timeout=60)

            if userinfo:
                self._non_session_authentications = {}

                for ui in userinfo["identity_set"]:
                    identity_id = ui["sub"]
                    last_authentication = ui.get("last_authentication")
                    last_auth_timestamp = 0

                    if last_authentication:
                        last_auth_timestamp = int(last_authentication)

                    self._non_session_authentications[identity_id] = last_auth_timestamp
        return self._non_session_authentications or {}

    @property
    def multi_factor_authentications(self):
        """
        Dictionary mapping Globus Auth Identity IDs to a boolean indicating
        whether the identity was authenticated in-session using multi-factor
        authentication. Defaults to 'False' for identities not in this dict.
        """
        return self._multi_factor_auth or {}

    def dependent_token(
        self,
        resource_server: Optional[str] = None,
        required_scope: Optional[str] = None,
    ) -> Optional[str]:
        """
        Obtain a dependent access token for the specified resource server and
        return it.

        :param resource_server: String name of the resource server.

        :returns Optional[str]:
            This function returns an access token as a string if available.
        """
        token = None
        rs_cache_key_format = "rs_dependent_{}_{}"
        scope_cache_key_format = "scope_dependent_{}_{}"
        now = int(time.time())

        if resource_server is None and required_scope is None or self.token is None:
            raise errors.InternalServerError("Invalid dependent token request")

        if resource_server:
            cache_key = rs_cache_key_format.format(self.token, resource_server)
            cached_token = cache.get(cache_key)
            if cached_token and cached_token["expires_at"] > now:
                if required_scope is None or required_scope in cached_token["scope"]:
                    return cached_token["access_token"]
        else:
            cache_key = scope_cache_key_format.format(self.token, required_scope)
            cached_token = cache.get(cache_key)
            if cached_token and cached_token["expires_at"] > now:
                if required_scope is None or required_scope in cached_token["scope"]:
                    return cached_token["access_token"]

        auth_client = current_app.auth_client
        try:
            tokens = auth_client.oauth2_get_dependent_tokens(self.token)
        except globus_sdk.AuthAPIError as exc:
            request_id = getattr(g, "request_id", None) if g else None
            raw_json = exc.raw_json or {}
            errs = raw_json.get("errors", [])
            auth_id = None
            for i in errs:
                if "id" in i:
                    auth_id = i["id"]
                    break

            current_app.logger.info(
                "auth_exception: gcs_request_id=%s "
                'auth_request_id=%s error=%s error_description="%s"',
                request_id,
                auth_id,
                raw_json.get("error"),
                raw_json.get("error_description"),
            )
            if required_scope:
                raise errors.MissingRequiredScopesError([required_scope])
            else:
                raise errors.UnauthorizedError(
                    "Unable to contact {}".format(resource_server)
                )

        for t in tokens.data:
            rs_cache_key = rs_cache_key_format.format(self.token, t["resource_server"])
            cache_exp = now + t["expires_in"]
            t["expires_at"] = cache_exp
            cache.set(rs_cache_key, t, timeout=cache_exp)

            scopes = t["scope"].split()
            for scope in scopes:
                scope_cache_key = scope_cache_key_format.format(self.token, scope)
                cache.set(scope_cache_key, t, timeout=cache_exp)

            if not token:
                if resource_server:
                    if t["resource_server"] == resource_server:
                        if required_scope is None or required_scope in scopes:
                            token = t["access_token"]
                else:
                    if required_scope in scopes:
                        token = t["access_token"]

        if token:
            return token
        elif required_scope:
            raise errors.MissingRequiredScopesError(scopes=[required_scope])
        else:
            raise errors.UnauthorizedError(
                "Unable to contact {}".format(resource_server)
            )
