from typing import Dict, List, TypeVar
import uuid

from flask import current_app
import globus_sdk

from globus.manager.api.authentication import AuthenticationState
import globus.manager.api.errors as errors
from .identity_in_identity_set_filter import IdentityInIdentitySetFilter
from .identity_set import IdentitySet
from globus.manager.api.models import Role
from globus.manager.constants import PrincipalType, RoleType
from globus.manager.api.utils.groups_client import GroupsClient


class RoleSpec:
    """
    This class parses the role specification from the RoleAuthorizer
    specification in the API documentation
    """

    def __init__(self, role_string, collection=None):
        assert role_string.startswith("collection:") or role_string.startswith(
            "endpoint:"
        )

        if role_string.startswith("collection:"):
            role, resource = role_string.rsplit(":", 1)
            assert resource in ["*", "{collection_id}"]
            if resource == "*":
                resource = None
            else:
                resource = collection.id
        else:
            role = role_string
            resource = None

        if resource is not None:
            assert uuid.UUID(resource) is not None

        self.role = RoleType(role)
        self.role_name = self.role.name.split("_", 1)[1]
        self.resource = resource
        self.collection = collection

    def __repr__(self):
        if (
            self.role == RoleType.endpoint_owner
            or self.role == RoleType.endpoint_administrator
            or self.role == RoleType.endpoint_activity_manager
            or self.role == RoleType.endpoint_activity_monitor
        ):
            return "<RoleSpec {}>".format(self.role.value)
        else:
            if self.resource:
                return "<RoleSpec {}:{}>".format(self.role.value, "{collection_id}")
            else:
                return "<RoleSpec {}:*>".format(self.role.value)

    def __str__(self):
        if (
            self.role == RoleType.endpoint_owner
            or self.role == RoleType.endpoint_administrator
            or self.role == RoleType.endpoint_activity_manager
            or self.role == RoleType.endpoint_activity_monitor
        ):
            return self.role.value
        else:
            if self.resource:
                return "{}:{}".format(self.role.value, self.collection)
            else:
                return "{}:*".format(self.role.value)


RoleSpecs = TypeVar("RoleSpecs", str, List[str], RoleSpec, List[RoleSpec])


class RoleAuthorizer:
    """
    This class provides methods to process role-based access control
    policies for endpoints or collections. Note that if there is no
    subscription on the endpoint, only the `endpoint:owner` role is
    considered.

    """

    def __init__(self, role_specs):
        """
        Initialize a RoleAuthorizer object.

        :param role_specs:
            The roles to authorize. This input may be in any of these forms:
            -   String version of one or more role names such as
                `endpoint:administrator` separated by the `,` character
            -   List of strings of role names such as `endpoint:administrator`
            -   A RoleSpec object
            -   A List of RoleSpec objects
        """
        if isinstance(role_specs, str):
            role_specs = role_specs.split(",")
        if isinstance(role_specs, RoleSpec):
            role_specs = [role_specs]
        self.role_specs = []
        for r in role_specs:
            if isinstance(r, RoleSpec):
                role_spec = r
            else:
                role_spec = RoleSpec(r)
            self.role_specs.append(role_spec)

    def authorize_identity(
        self, identity_id: str, auth_state: AuthenticationState
    ) -> dict:
        """
        Attempt to authorize the caller as identity_id (a string UUID)
        using the authorization information in the current request's
        authentication state.

        :param identity_id:
            A string representation of the uuid to authorize.
        :param auth_state:
            The caller's authentication state. This is passed in rather
            than based on `g` to handle the gridftp case where the
            authentication state is created based on assertions from transfer.

        :returns:
            This function returns an entry from the auth_state's
            identity_set_data property that contains the information
            about the identity which was authorized, with two additional
            entries:

            -   _role_name: A string containing the name of the role
                the caller has, such as `endpoint:administrator`
            -   _role_id: A string representation of the UUID of
                this role.

        :raises errors.IdNotInIdentitySetError:
            The `auth_state` does not have any identity which matches
            `identity_id`.
        :raises errors.MissingRequiredRoleError:
            The `auth_state` has the given identity, but there is no
            role assignment for that identity.
        """
        idset = IdentitySet(identity_set=auth_state.identity_set_data)

        current_app.logger.debug(
            "Filtering identity_set %s with %s", idset, identity_id
        )

        # Filter out identities which don't match the requested id first
        f = IdentityInIdentitySetFilter(identity_id)
        f(idset, auth_state)

        if not idset:
            raise errors.IdNotInIdentitySetError(detail=idset.error_details[0])

        if current_app.subscription:
            enforced_roles = self.role_specs
        else:
            enforced_roles = [
                r for r in self.role_specs if r.role == RoleType.endpoint_owner
            ]

        identity = idset[0]
        for role in Role.query.filter(
            ~Role.deleted,  # type: ignore
            Role.principal_type == PrincipalType.identity.value,
            Role.identity_id == identity["id"],
            Role.role.in_(set([r.role.name for r in enforced_roles])),
        ).all():
            for spec in enforced_roles:
                if role.role == spec.role and (
                    spec.resource is None or role.resource == spec.resource
                ):
                    res_identity = identity.copy()
                    res_identity["_role_name"] = spec.role.value
                    res_identity["_role_id"] = role.id
                    return res_identity

        # Look up group roles if any and store them in a list in
        # group_roles keyed by the group id
        group_roles = {}  # type: Dict[str, List[Role]]
        for role in Role.query.filter(
            ~Role.deleted,  # type: ignore
            Role.principal_type == PrincipalType.group,
            Role.role.in_(set([r.role.name for r in enforced_roles])),
        ).all():
            for spec in enforced_roles:
                if role.role == spec.role and (
                    spec.resource is None or role.resource == spec.resource
                ):
                    if role.identity_id not in group_roles:
                        group_roles[role.identity_id] = []
                    group_roles[role.identity_id].append(role)

        # Look up those group IDs using the new my_groups API, and
        # determine if the membership matches the specified identity
        if group_roles:
            groups_response = None
            try:
                groups_token = auth_state.dependent_token(
                    required_scope="urn:globus:auth:scope:"
                    "groups.api.globus.org:view_my_groups_and_memberships"
                )
                groups_client = GroupsClient(
                    authorizer=globus_sdk.AccessTokenAuthorizer(groups_token),
                    environment=current_app.environment,
                )
                groups_response = groups_client.get_my_groups()
                for my_groups in groups_response:
                    for role in group_roles.get(my_groups["id"], []):
                        for membership in my_groups["my_memberships"]:
                            assert membership["group_id"] == my_groups["id"]
                            # TODO: does enforce_authentication_timeout matter
                            # here?
                            if identity["id"] == membership["identity_id"]:
                                res_identity = identity.copy()
                                res_identity["_role_name"] = role.role.value
                                res_identity["_role_id"] = role.id
                                return res_identity
            except Exception as e:
                current_app.logger.warning(
                    "Group resolution failure: %s %s %s",
                    group_roles,
                    groups_response,
                    e,
                )
                pass

        raise errors.MissingRequiredRoleError(roles=enforced_roles)

    def authorize_identities(self, auth_state: AuthenticationState) -> IdentitySet:
        """
        Attempt to authorize the caller based on the roles associated
        with this authorizer.

        :param auth_state:
            The caller's authentication state. This is passed in rather
            than based on `g` to handle the gridftp case where the
            authentication state is created based on assertions from transfer.

        :returns:
            This function returns an IdentitySet containing the list of
            the caller's identities which are authorized by this
            role authorizer. This IdentitySet may be empty if no role
            is assigned for this identity_set. In that case
            IdentitySet.error_details may include information about why
            it is empty.
        """
        identities = IdentitySet()
        details = []
        for i in auth_state.identity_set_data:
            try:
                identity = self.authorize_identity(i["id"], auth_state)
                identities.append(identity)
            except errors.APIBaseError as e:
                current_app.logger.debug("Exception (%s) %s", type(e), e)
                if e.detail:
                    details.append(e.detail)
        if not identities:
            identities.error_details.extend(details)

        return identities

    def __repr__(self):
        return "<RoleAuthorizer {}>".format(self.role_specs)

    def __str__(self):
        return "role_authorizer allowed_roles={}".format(
            ",".join(str(s) for s in self.role_specs)
        )


EndpointRoleAuthorizer = RoleAuthorizer(
    [
        "endpoint:owner",
        "endpoint:administrator",
        "endpoint:activity_manager",
        "endpoint:activity_monitor",
    ]
)
