from typing import List

from globus.manager.constants import ConfigurationObjectTag
from .user_credential import UserCredential
from .decorators import checks_policy_property, checks_property


class BlackPearlUserCredential(UserCredential):
    POLICY_DATA_TYPE = "blackpearl_user_credential_policies#1.0.0"
    POLICY_DATA_TYPES = [POLICY_DATA_TYPE]
    CONNECTOR_ID = "7e3f3f5e-350c-4717-891a-2f451c24b0d4"

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.USER_CREDENTIAL.name + CONNECTOR_ID
        )
    }

    @property
    def connector_properties(self) -> List[str]:
        return [
            "access_id",
            "secret_key",
        ]

    def __init__(self, appsync_data=None, doc=None, storage_gateway=None):
        if doc:
            username = doc["username"]
            access_id, secret_key = storage_gateway.get_access_id_info(username)
            policies = doc.get("policies")
            if policies is None:
                policies = {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                    "access_id": access_id,
                    "secret_key": secret_key,
                }
            else:
                # This should have been removed by the deserializer, and
                # so we'll override with the correct values
                policies["access_id"] = access_id
                policies["secret_key"] = secret_key

            doc = {
                **doc,
                **{
                    "display_name": username,
                    "invalid": doc.get("invalid", False),
                    "deleted": False,
                    "provisioned": True,
                    "policies": policies,
                },
            }

        super().__init__(
            appsync_data=appsync_data, doc=doc, storage_gateway=storage_gateway
        )

    @property
    def connector_id(self):
        return self.CONNECTOR_ID

    @connector_id.setter
    def connector_id(self, value):
        pass

    @property
    def access_id(self):
        value = self.storage_gateway.get_access_id(self.username)
        if not value:
            raise AttributeError("access_id")
        return value

    @access_id.setter
    def access_id(self, value):
        pass

    @access_id.deleter
    def access_id(self):
        pass

    @checks_policy_property(prop="access_id")
    def check_access_id(self, value: str):
        current = getattr(self, "access_id", None)
        if current and value and current != value:
            raise ValueError("access_id")

    @property
    def secret_key(self):
        value = self.storage_gateway.get_secret_key(self.username)
        if not value:
            raise AttributeError("secret_key")
        return value

    @secret_key.setter
    def secret_key(self, value):
        pass

    @secret_key.deleter
    def secret_key(self):
        pass

    @checks_policy_property(prop="secret_key")
    def check_secret_key(self, value: str):
        current = getattr(self, "secret_key", None)
        if current and value and current != value:
            raise ValueError("secret_key")

    @checks_property(prop="display_name")
    def check_policy_property(self, value: str):
        if getattr(self, "display_name", None) and self.display_name != value:
            raise AttributeError()

    def patch(self, **doc):
        # display_name is computed from user_name, so disallow changes
        display_name = doc.pop("display_name", None)
        if not display_name:
            display_name = self.username
        doc["display_name"] = display_name

        policies = doc.pop("policies", {})
        if not policies:
            policies["DATA_TYPE"] = self.POLICY_DATA_TYPE
        doc["policies"] = policies
        super().patch(**doc)

    def put(self, **doc):
        # display_name is computed from user_name, so disallow changes
        display_name = doc.get("display_name")
        if not display_name:
            display_name = self.username

        doc["display_name"] = display_name

        policies = doc.pop("policies", {})
        if not policies:
            policies["DATA_TYPE"] = self.POLICY_DATA_TYPE
        doc["policies"] = policies
        super().put(**doc)

    def __repr__(self):
        return "<BlackPearlUserCredential {}>".format(self.id)
