from typing import Dict
import uuid

import marshmallow as ma
from marshmallow_oneofschema import OneOfSchema

from .data_type import DataTypeSchema, data_type_schema
from .identity_mapping import IdentityMappingSchema
from ..patch_schema import create_patch_schema
from .path_restrictions import path_restrictions_schema
from .schema_version_checker import verify_schema_version, version_safe_type_schemas


class StorageGatewaySchema_1_0_0(ma.Schema):
    """
    Storage Gateway
    """

    _DATA_TYPE = "storage_gateway#1.0.0"
    DATA_TYPE = ma.fields.Str(default=_DATA_TYPE)
    id = ma.fields.UUID()
    display_name = ma.fields.String(max=256)
    connector_id = ma.fields.UUID()
    high_assurance = ma.fields.Boolean(required=False, allow_none=True)
    require_high_assurance = ma.fields.Boolean(required=False, allow_none=True)
    authentication_assurance_timeout = ma.fields.Int(
        required=False,
        allow_none=True,
        validate=[ma.validate.Range(min=1)],
        load_only=True,
    )
    authentication_timeout_mins = ma.fields.Int(
        required=False, allow_none=True, validate=[ma.validate.Range(min=1)]
    )
    allowed_domains = ma.fields.List(ma.fields.String)
    identity_mappings = ma.fields.List(
        ma.fields.Nested(IdentityMappingSchema), required=False, allow_none=True
    )
    users_allow = ma.fields.List(ma.fields.Str, required=False, allow_none=True)
    users_deny = ma.fields.List(ma.fields.Str, required=False, allow_none=True)
    restrict_paths = ma.fields.Nested(
        path_restrictions_schema, required=False, allow_none=True
    )
    policies = ma.fields.Method(serialize="dump_policies", deserialize="load_policies")
    process_user = ma.fields.Str(required=False, allow_none=True)
    load_dsi_module = ma.fields.Str(required=False, allow_none=True)
    deleted = ma.fields.Boolean(required=False)

    def load_policies(self, value):
        return data_type_schema.load(value, partial=self.partial)

    def dump_policies(self, obj):
        connector_id = str(ma.utils.get_value(obj, "connector_id"))
        if not self.load_only:
            schema = self.connector_policies_private_schemas.get(connector_id)
        else:
            schema = self.connector_policies_schemas.get(connector_id)
        if schema:
            return schema.dump(obj)
        else:
            raise ValueError("Unable to serialize {} {}".format(obj, connector_id))

    @ma.validates("DATA_TYPE")
    def validate_data_type(self, data, **kwargs):
        verify_schema_version(data, self._DATA_TYPE)

    @ma.post_load
    def backward_compatible_assurance_timeout_load(self, obj, **kwargs):
        authentication_assurance_timeout = obj.pop(
            "authentication_assurance_timeout", None
        )
        if authentication_assurance_timeout:
            obj["authentication_timeout_mins"] = authentication_assurance_timeout
        return obj

    @ma.post_dump
    def backward_compatible_require_high_assurance_dump(self, data, **kwargs):
        data["require_high_assurance"] = data.get("high_assurance")
        return data

    @ma.post_load
    def backward_compatible_require_high_assurance_load(self, obj, **kwargs):
        if "high_assurance" not in obj and "require_high_assurance" in obj:
            obj["high_assurance"] = obj.pop("require_high_assurance")
        return obj

    @ma.post_load
    def stringify_uuid(self, obj, **kwargs):
        res = {}
        for k, v in obj.items():
            if isinstance(v, uuid.UUID):
                res[k] = str(v)
            elif isinstance(v, list):
                newlist = []
                for i in v:
                    if isinstance(i, uuid.UUID):
                        newlist.append(str(i))
                    else:
                        newlist.append(i)
                res[k] = newlist
            else:
                res[k] = v
        return res

    # connectors may register their schema types in this dict, with
    # the connector ID as the key
    connector_policies_schemas = {}  # type: Dict[str, ma.Schema]
    connector_policies_private_schemas = {}  # type: Dict[str, ma.Schema]
    private = [
        "deleted",
        "identity_mappings",
        "restrict_paths",
        "process_user",
        "load_dsi_module",
        "users_allow",
        "users_deny",
        "authentication_assurance_timeout",  # private since it is deprecated
    ]

    @classmethod
    def register_schema(cls, public=None, private=None):
        if public:
            connector_id = public.connector_id
            cls.connector_policies_schemas[connector_id] = public
        if private:
            connector_id = private.connector_id
            cls.connector_policies_private_schemas[connector_id] = private


class StorageGatewaySchema_1_1_0(StorageGatewaySchema_1_0_0):
    _DATA_TYPE = "storage_gateway#1.1.0"
    DATA_TYPE = ma.fields.Str(default=_DATA_TYPE)

    require_mfa = ma.fields.Boolean(required=False, default=False)


storage_gateway_1_0_0_schema = StorageGatewaySchema_1_0_0(
    load_only=StorageGatewaySchema_1_0_0.private
)
storage_gateway_1_0_0_private_schema = StorageGatewaySchema_1_0_0()
storage_gateway_1_0_0_patch_schema = create_patch_schema(StorageGatewaySchema_1_0_0)

storage_gateway_1_1_0_schema = StorageGatewaySchema_1_1_0(
    load_only=StorageGatewaySchema_1_1_0.private
)
storage_gateway_1_1_0_private_schema = StorageGatewaySchema_1_1_0()
storage_gateway_1_1_0_patch_schema = create_patch_schema(StorageGatewaySchema_1_1_0)


class StorageGatewaySchema(OneOfSchema):
    _DATA_TYPE = storage_gateway_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = version_safe_type_schemas(
        {
            "storage_gateway#1.0.0": storage_gateway_1_0_0_schema,
            "storage_gateway#1.1.0": storage_gateway_1_1_0_schema,
        }
    )
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE

    @classmethod
    def register_schema(cls, public=None, private=None):
        for cls in cls.type_schemas.values():
            cls.register_schema(public=public, private=private)


class StorageGatewayPrivateSchema(OneOfSchema):
    _DATA_TYPE = storage_gateway_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "storage_gateway#1.0.0": storage_gateway_1_0_0_private_schema,
        "storage_gateway#1.1.0": storage_gateway_1_1_0_private_schema,
    }
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE


class StorageGatewayPatchSchema(OneOfSchema):
    _DATA_TYPE = storage_gateway_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = version_safe_type_schemas(
        {
            "storage_gateway#1.0.0": storage_gateway_1_0_0_patch_schema,
            "storage_gateway#1.1.0": storage_gateway_1_1_0_patch_schema,
        }
    )
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE


storage_gateway_schema = StorageGatewaySchema(
    load_only=StorageGatewaySchema_1_0_0.private
)
storage_gateway_private_schema = StorageGatewayPrivateSchema()
storage_gateway_patch_schema = create_patch_schema(StorageGatewayPatchSchema)

DataTypeSchema.register_schema(storage_gateway_1_0_0_schema)
DataTypeSchema.register_schema(storage_gateway_1_1_0_schema)
