from typing import List

import marshmallow as ma

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


AZURE_BLOB_CONNECTOR_ID = "9436da0c-a444-11eb-af93-12704e0d6a4d"


class AzureBlobStoragePoliciesSchema(ma.Schema):
    _DATA_TYPE = "azure_blob_storage_policies#1.0.0"
    DATA_TYPE = ma.fields.Str(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    auth_type = ma.fields.Str(required=True, allow_none=True)
    client_id = ma.fields.Str(required=True, allow_none=True)
    secret = ma.fields.Str(required=True, allow_none=True)
    tenant = ma.fields.Str(required=True, allow_none=True)
    account = ma.fields.Str(required=True, allow_none=True)
    adls = ma.fields.Boolean(required=True, allow_none=True)
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        doc="URL of the auth callback that must be registered on the Microsoft "
        "API console for the application client_id in order to process "
        "Microsoft credentials.",
    )

    private = ["client_id", "secret", "tenant", "account", "adls"]
    connector_id = AZURE_BLOB_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data


class AzureBlobCollectionPoliciesSchema(ma.Schema):
    _DATA_TYPE = "azure_blob_collection_policies#1.0.0"
    DATA_TYPE = ma.fields.String(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    private = []  # type: List[str]
    connector_id = AZURE_BLOB_CONNECTOR_ID


class AzureBlobUserCredentialPoliciesSchema(ma.Schema):
    _DATA_TYPE = "azure_blob_user_credential_policies#1.0.0"
    DATA_TYPE = ma.fields.Str(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    sub = ma.fields.Str(allow_none=True)
    email = ma.fields.Email(allow_none=True)
    access_token = ma.fields.Str(required=False, allow_none=True)
    refresh_token = ma.fields.Str(required=False, allow_none=True)
    token_expiry = ma.fields.AwareDateTime()
    scopes = ma.fields.List(ma.fields.Str)
    tid = ma.fields.Str()

    private = ["access_token", "refresh_token"]
    connector_id = AZURE_BLOB_CONNECTOR_ID


azure_blob_user_credential_policies_private_schema = (
    AzureBlobUserCredentialPoliciesSchema()
)
azure_blob_user_credential_policies_schema = AzureBlobUserCredentialPoliciesSchema(
    load_only=AzureBlobUserCredentialPoliciesSchema.private
)
azure_blob_storage_policies_private_schema = AzureBlobStoragePoliciesSchema()
azure_blob_storage_policies_schema = AzureBlobStoragePoliciesSchema(
    load_only=AzureBlobStoragePoliciesSchema.private
)
azure_blob_collection_policies_schema = AzureBlobCollectionPoliciesSchema(
    load_only=AzureBlobCollectionPoliciesSchema.private
)
azure_blob_collection_policies_private_schema = AzureBlobCollectionPoliciesSchema()

DataTypeSchema.register_schema(azure_blob_storage_policies_schema)
DataTypeSchema.register_schema(azure_blob_user_credential_policies_schema)
DataTypeSchema.register_schema(azure_blob_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=azure_blob_storage_policies_schema,
    private=azure_blob_storage_policies_private_schema,
)

UserCredentialSchema.register_schema(
    public=azure_blob_user_credential_policies_schema,
    private=azure_blob_user_credential_policies_private_schema,
)

CollectionSchema.register_schema(
    public=azure_blob_collection_policies_schema,
    private=azure_blob_collection_policies_private_schema,
)
