"""
globus-connect-server endpoint migrate4
"""
import os
import pwd
import json
import click
import textwrap
import marshmallow as ma
from flask import current_app
from typing import List

import globus_sdk
import gcs_migration.check as check
from gcs_migration.util import do_link_auth_flow, get_api
import globus.manager.api.models as models
import globus.manager.api.schemas.api as schema
from gcs_cli.parsing.command_decorator import globus_command
from gcs_migration.models import MigrationPlan
from gcs_migration.commands.login import check_logged_in
from globus.manager.appsync.query import GraphQLError
from globus.manager.constants import CollectionType, RoleType
from globus.manager.api.blueprints.collections import allocate_domain
from globus.manager.api.app import make_app
from globus.manager.utils.keychain import NodeKey
from globus.manager.scripts.utils.setup_utils import load_keychain
from globus.manager.api.blueprints.utils import (
    create_object,
    delete_object,
    patch_object,
    update_object,
)


@globus_command(
    "migrate4",
    short_help="Migrate a Globus Connect Server v4",
    help="Migrate a Globus Connect Server v4 endpoint for use with v5.4",
    disable_options=["format", "explicit_host"],
)
@click.option("--verbose", "-v", is_flag=True)
@click.option("--finalize", "-f", is_flag=True)
def migrate4(verbose: bool, finalize: bool) -> None:
    # Run as root to get access to our log file.
    if os.getuid() != 0:
        click.echo("This command must be run as the root user", err=True)
        click.get_current_context().exit(1)

    if verbose:
        os.environ["GCS_MANAGER_LOG_LEVEL"] = "DEBUG"

    app = make_app(app_config="globus.manager.api.config.endpoint_migrate4")
    models.db.init_app(app)

    app.logger.info("migrate4: start")

    with app.app_context():
        models.db.create_all()
        models.db.session.commit()

        for conf in models.Configuration.query:
            if conf.name == "node_key":
                value = NodeKey(**json.loads(conf.value))
            else:
                value = conf.value
            app.config[conf.name.upper()] = value

        if "CLIENT_ID" not in app.config:
            click.echo("This node has not been set up.", err=True)
            click.get_current_context().exit(1)

        # Load the keychain
        load_keychain()

        migrate4 = Migrate4()
        migrate4.run()

        if finalize:
            if not check_logged_in():
                do_link_auth_flow(
                    session_params={"session_required_single_domain": "globusid.org"}
                )
            transfer_client = get_api()
            migrate4.finalize(transfer_client)


class Migrate4:
    CONNECTOR_MAP = {
        "7251f6c8-93c9-11eb-95ba-12704e0d6a4d": "ActiveScale",
        "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63": "Box",
        "1b6374b0-f6a4-4cf7-a26f-f262d9c6ca72": "Ceph",
        "56366b96-ac98-11e9-abac-9cb6d0d9fd63": "Google Cloud Storage",
        "976cf0cf-78c3-4aab-82d2-7c16adbcc281": "Google Drive",
        "e47b6920-ff57-11ea-8aaa-000c297ab3c2": "iRODS",
        "28ef55da-1f97-11eb-bdfd-12704e0d6a4d": "OneDrive",
        "145812c8-decc-41f1-83cf-bb2a85a2a70b": "POSIX",
        "052be037-7dda-4d20-b163-3077314dc3e6": "POSIX Staging",
        "7643e831-5f6c-4b47-a07f-8ee90f401d23": "S3",
        "7e3f3f5e-350c-4717-891a-2f451c24b0d4": "SpectraLogic BlackPearl",
    }
    SUPPORTED_CONNECTORS = ["POSIX"]

    COLLECTION_DISPLAY_NAME_FORMAT = "Migration: {}"
    COLLECTION_KEYWORD_PREFIX = "migration_"
    COLLECTION_KEYWORD_FORMAT = COLLECTION_KEYWORD_PREFIX + "{}"

    def run(self):
        tasks = [
            self._load_migration_plan,
            self._check_subscription,
            self._verify_support_for_plan,
            self._verify_connector_properties,
            self._execute_migration_plan,
        ]

        for task in tasks:
            task()

    def finalize(self, transfer_client: globus_sdk.TransferClient) -> None:
        """
        Finalize migration by transferring the v4 collection's id to the v5
        endpoint object in transfer. This involves some low-level changes to
        the models to keep them coherent.
        """
        mapped_v4_id = self.plan_docs.mapped_collection.v4_collection_id

        # IDs in this list are collections where there is already a model
        # in the local database which matches the v4 collection ID. Any id
        # in this list does not need to be migrated.
        migrated_and_in_db_ids: List[str] = []

        # IDs in this list are collections where the transfer endpoint for
        # the collection has been migrated to the new uuid, but we don't
        # yet have a GCS model for the collection, we'll need to reconstruct
        # the model and store it in appsync/local db
        migrated_without_model_collection_ids: List[str] = []

        # Pass 1: Set the gcs_private_temporary_migration_id value on each
        # of the v4 endpoint documents to allow the new v5 collection to
        # take ownership of its uuid. This is done with the token from the
        # owner of the v4 endpoint.
        for collection in [
            self.plan_docs.mapped_collection
        ] + self.plan_docs.guest_collections:
            already_migrated_model = models.Collection.query.filter(
                models.Collection.id == collection.v4_collection_id,
                ~models.Collection.deleted,  # type: ignore
            ).one_or_none()
            if already_migrated_model:
                # If there is already a collection in the db with the v4 id,
                # then assume the uuid migration has already occurred for this
                # collection, ad add it to the migrated_and_in_db_ids list
                migrated_and_in_db_ids.append(already_migrated_model.id)
                continue
            elif collection.v4_collection_id == collection.v5_collection_id:
                # If the transfer object with the migration id keyword already
                # matches the persistent ID, then we've already completed the
                # migration in transfer and need to deal with an unsynchronized
                # model.
                migrated_without_model_collection_ids.append(
                    collection.v4_collection_id
                )
                continue
            elif getattr(collection, "disable", False):
                continue

            # temporary_collection_model is the GCS model for the collection
            # before the uuid was migrated.
            temporary_collection_model: models.Collection = (
                models.Collection.query.filter(
                    models.Collection.id == collection.v5_collection_id
                ).one_or_none()
            )

            if temporary_collection_model is None:
                raise click.ClickException(
                    f"Attempting to migrate collection {collection.v5_collection_id} "
                    "but it is not present in the database."
                )

            try:
                # Set the flag in the transfer endpoint to allow this
                # endpoint's id to be transferred to the v5 collection's
                # endpoint
                transfer_client.update_endpoint(
                    collection.v4_collection_id,
                    {"gcs_private_temporary_migration_id": collection.v5_collection_id},
                )
            except globus_sdk.exc.GlobusAPIError as e:
                # This could happen if the migration was already done for the
                # collection, but the models weren't committed to the DB.
                if e.http_status == 403:
                    self._is_collection_owner_equal(
                        transfer_client,
                        collection.v4_collection_id,
                        current_app.client_id,
                    )
                    migrated_without_model_collection_ids.append(
                        collection.v4_collection_id
                    )
                else:
                    raise click.ClickException(
                        f"Unable to migrate collection {collection.v4_collection_id}"
                    )

        # If we think the mapped collection has already been migrated, find its
        # model so we can use it as the mapped_collection when migrating the
        # other collections; otherwise we'll set it to None for now. In that
        # case, we'll set it to the model after migrating it.
        if mapped_v4_id in migrated_and_in_db_ids:
            mapped_collection_model = models.Collection.query.filter(
                models.Collection.id == mapped_v4_id
            ).one()
        else:
            mapped_collection_model = None

        # Pass 2:
        # * Update the v5 collections using the private API to take
        #   ownership of the v4 endpoint UUIDs. This returns a new collection
        #   object which we will add to appsync, it also destroys the v5
        #   temporary collection in transfer, so we'll have an object that
        #   refers to stale data if we are terminated before committing the
        #   changes to the database.
        # * Update the roles on the v5 collection be re-parented to the
        #   collection with the persistent collection uuid.
        for collection in [
            self.plan_docs.mapped_collection
        ] + self.plan_docs.guest_collections:
            if getattr(collection, "disable", False):
                continue
            elif collection.v4_collection_id not in migrated_and_in_db_ids:
                temporary_collection_model = models.Collection.query.filter(
                    models.Collection.id == collection.v5_collection_id
                ).one_or_none()
                # The model with the temporary id is not present, but we
                # don't think the collection is migrated. This is an
                # inconsistent state
                if temporary_collection_model is None:
                    current_app.logger.error(
                        "Inconsistent state, temporary collection "
                        f"{collection.v5_collection_id} is not present"
                    )
                    raise click.ClickException(
                        f"Collection {collection.v4_collection_id} in inconsistent state"
                    )

                if (
                    collection.v4_collection_id
                    not in migrated_without_model_collection_ids
                ):
                    try:
                        current_app.transfer_client.post(
                            f"/private/endpoint/{collection.v5_collection_id}/"
                            f"migrate_id?persistent_endpoint_id={collection.v4_collection_id}"
                        )
                    except Exception as e:
                        current_app.logger.error(
                            f"Error migrating endpoint id {collection.v4_collection_id} "
                            f"to collection {collection.v5_collection_id}: {e}"
                        )
                        raise click.ClickException("Unable to migrate collection id")

                domain_name = allocate_domain(collection.doc["collection_type"])
                temp_display_name = temporary_collection_model.display_name
                temp_prefix = self.COLLECTION_DISPLAY_NAME_FORMAT.format("")
                temporary_collection_model.display_name = temp_display_name.replace(
                    temp_prefix, "", 1
                )

                # Collection.migrate_collection_uuid() will return a new
                # model that uses the persistent id, and marks the
                # temporary_collection_model as deleted.
                final_model = temporary_collection_model.migrate_collection_uuid(
                    collection.v4_collection_id,
                    domain_name,
                    mapped_collection=mapped_collection_model,
                )
                # If we migrated the mapped collection, track the model for
                # other guest migrations
                if final_model.is_mapped_collection:
                    mapped_collection_model = final_model

                try:
                    # Add the new model to appsync
                    final_model = current_app.appsync_client.create_object(final_model)
                except GraphQLError as e:
                    if e.errors[0].get("errorType") == "CONFLICT":
                        if e.data is None:
                            raise click.ClickException(
                                f"Sync conflict for {collection.v4_collection_id} {e}"
                            )
                        # The model already exists in appsync from a previous
                        # migration attempt. Overwrite that with our model so we
                        # don't need to replace the transfer document and redo the
                        # scope text.
                        final_model.serial = e.data["items"][0]["serial"]
                        try:
                            final_model = current_app.appsync_client.update_object(
                                final_model
                            )
                        except GraphQLError as e:
                            raise click.ClickException(
                                f"Sync problem for {collection.v4_collection_id} {e}"
                            )

                models.db.session.add(final_model)
                models.db.session.commit()
            else:
                final_model = models.Collection.query.filter(
                    models.Collection.id == collection.v4_collection_id
                ).one_or_none()

            if not final_model:
                current_app.logger.error(
                    "Migration in inconsistent state, collection model for "
                    f"{collection.v4_collection_id} does not exist"
                )
                raise click.ClickException("Invalid state")

            for role in models.Role.query.filter(
                models.Role.resource == collection.v5_collection_id,
                ~models.Role.deleted,  # type: ignore
            ):
                # Move roles to the final model
                try:
                    update_object(
                        role, models.Role.migrate_collection_uuid, final_model
                    )
                except Exception as e:
                    current_app.logger.error(
                        "Could not update role object for collection "
                        f"{collection.v4_collection_id}: {e}"
                    )
                    raise click.ClickException(
                        f"Migration unable to update role {role.id} for collection "
                        f"{collection.v4_collection_id}"
                    )

        # Pass 3:
        # Delete the collection objects which use the temporary v5 UUIDs. Ordered
        # in reverse (guest before mapped so we can hard-delete things)
        for collection in self.plan_docs.guest_collections + [
            self.plan_docs.mapped_collection
        ]:
            if collection.v5_collection_id != collection.v4_collection_id:
                model = models.Collection.query.filter(
                    models.Collection.id == collection.v5_collection_id
                ).one_or_none()
                if model:
                    try:
                        model.unregister_child_client()
                    except Exception as e:
                        current_app.logger.error(
                            "Error deleting temporary collection model for collection "
                            f"{collection.v5_collection_id}: {e}"
                        )
                        raise click.ClickException("Error cleaning up data")
                    models.ConfigurationObject.delete(model)
                    try:
                        current_app.appsync_client.delete_object(model)
                    except GraphQLError as e:
                        if e.errors[0].get("errorType") != "CONFLICT":
                            current_app.logger.error(
                                "Error deleting temporary collection model for collection "
                                f"{collection.v5_collection_id}: {e}"
                            )
                            raise click.ClickException(
                                "Error synchronizing deleted collection"
                            )
                    models.db.session.delete(model)
                    models.db.session.commit()

    def _is_collection_owner_equal(
        self,
        transfer_client: globus_sdk.TransferClient,
        collection_id: str,
        owner_id: str,
    ) -> bool:
        try:
            res = transfer_client.get_endpoint(collection_id)
            if res["owner_id"] == owner_id:
                current_app.logger.info(
                    f"Collection {collection_id} is already "
                    "migrated in the Globus Transfer"
                )
                return True
            else:
                current_app.logger.error(
                    f"Collection {collection_id} can not be migrated, it is "
                    f"owned by globus user {res['owner_id']}"
                )
                raise click.ClickException(
                    "Unable to migrate collection {collection_id}"
                )

        except Exception as e:
            current_app.logger.info(
                f"Unable to to prepare endpoint {collection_id} for migration. "
                f"The endpoint is not accessible in the transfer service: {e}"
            )
            raise click.ClickException(f"Unable to migrate collection {collection_id}")

    def _load_migration_plan(self):
        """
        Load the migration plan from disk and verify it.

        The important instance variables setup here and used in the remainder
        of this class are:

        - self.imported_plan is a gcs_migration.MigrationPlan
        - self.plan_docs is an instance of MigrationPlanDocs
        """
        click.echo("Loading the migration plan from disk.")

        # Load from disk
        self.imported_plan = _import_migration_plan()
        # Run internal gcs_migration checks
        _check_imported_plan(self.imported_plan)
        # No support for cross-environment migration
        self._check_globus_environment(
            current_app.environment, self.imported_plan.environment
        )
        # Create and verify the component docs for the endpoint sync
        self.plan_docs = MigrationPlanDocs(self.imported_plan)

    def _check_subscription(self):
        """
        Check for a subscription to cover paid features
        """
        click.echo("Checking subscription")
        sub_id = current_app.subscription

        # We assume any subscription covers all features
        if sub_id is not None:
            return

        # Connector ID
        connector_id = self.plan_docs.storage_gateway.doc["connector_id"]
        if self.CONNECTOR_MAP.get(connector_id, None) != "POSIX":
            click.echo(
                "The migration plan includes a connector that requires a subscription. "
                "Set a subscription on this endpoint and rerun this command.",
                err=True,
            )
            click.get_current_context().exit(0)

        # Guest collections
        if self.plan_docs.guest_collections:
            click.echo(
                "The migration plan includes guest collections which requires a subscription. "
                "Set a subscription on this endpoint and rerun this command.",
                err=True,
            )
            click.get_current_context().exit(0)

        collection = self.plan_docs.mapped_collection
        if len(collection.roles) > 1:
            click.echo(
                "The migration plan includes roles which require a subscription. "
                "Set a subscription on this endpoint and rerun this command.",
                err=True,
            )
            click.get_current_context().exit(0)

        if collection.doc.get("enable_https"):
            click.echo(
                "The migration plan includes the https interface which requires a subscription. "
                "Set a subscription on this endpoint and rerun this command.",
                err=True,
            )
            click.get_current_context().exit(0)

    def _verify_support_for_plan(self):
        """
        Check the migration plan doesn't ask for anything we don't yet support.
        """
        click.echo("Checking the migration plan for unsupported features.")

        storage_gateway = self.plan_docs.storage_gateway
        # No support for HA gateways
        if storage_gateway.doc.get("high_assurance", False) is True:
            click.echo(
                "High assurance gateways are not supported with migration.", err=True
            )
            click.get_current_context().exit(1)

        # Make sure it is a known connector_id
        connector_id = storage_gateway.doc["connector_id"]
        connector_type = self.CONNECTOR_MAP.get(connector_id)

        if connector_type is None:
            click.echo(
                f"Unknown connector id {storage_gateway.doc['connector_id']}.", err=True
            )
            click.get_current_context().exit(1)

        # Make sure it is a supported connector_id
        if connector_type not in self.SUPPORTED_CONNECTORS:
            click.echo(
                f"{connector_type} connector is not yet supported for migration.",
                err=True,
            )
            click.get_current_context().exit(1)

    def _verify_posix_connector_properties(self):
        """
        Verify anything about the migration plan that is specific to POSIX.

        We need to verify that usernames in guest collection user credentials are
        valid local usernames.
        """
        click.echo("Verifying properties specific to the POSIX connector.")

        local_usernames = []
        for guest in self.plan_docs.guest_collections or []:
            if guest.credential.doc["username"] not in local_usernames:
                local_usernames.append(guest.credential.doc["username"])

        missing_usernames = []
        for username in local_usernames:
            try:
                pwd.getpwnam(username)
            except KeyError:
                missing_usernames.append(username)

        if missing_usernames:
            click.echo(
                "\n".join(
                    textwrap.wrap(
                        "These POSIX usernames are not available on this system. Please "
                        "add these accounts to this node and rerun this command."
                    )
                    + [""]
                    + textwrap.wrap(f"Missing accounts: {', '.join(missing_usernames)}")
                ),
                err=True,
            )
            click.get_current_context().exit(1)

    def _verify_connector_properties(self):
        switcher = {
            "145812c8-decc-41f1-83cf-bb2a85a2a70b": self._verify_posix_connector_properties,
        }

        func = switcher.get(self.plan_docs.storage_gateway.doc["connector_id"])
        if func:
            func()

    def _execute_migration_plan(self):
        click.echo("Executing the migration plan.")

        gateway = self._sync_storage_gateway(self.plan_docs.storage_gateway)
        mapped = self._sync_mapped_collection(self.plan_docs.mapped_collection, gateway)
        self._sync_guest_collections(self.plan_docs.guest_collections, gateway, mapped)

    def _sync_storage_gateway(self, plan):
        """
        Create or update the storage gateway

        We name the storage gateway according to the migration plan, without any special
        formatting since it is not a public object and there is no corresponding v4 version.

        We expect this display name to remain unchange so that we can look it up on successive
        iterations of this command. If the gateway display_name is changed, we assume it does
        not exist and we recreate it (and everything below it).
        """
        click.echo("Syncing the storage gateway")
        gateway = self._lookup_storage_gateway(plan.doc["display_name"])

        if gateway is None:
            click.echo(".Creating the new v5 storage gateway")
            return self._create_storage_gateway(plan.doc)

        update_doc = self._create_gateway_update_doc(plan.doc, gateway)
        if update_doc:
            click.echo(".Updating the existing v5 storage gateway")
            self._update_storage_gateway(update_doc, gateway)
        return gateway

    def _lookup_storage_gateway(self, display_name):
        current_app.logger.debug(f"Checking if storage gateway '{display_name}' exists")
        return models.StorageGateway.query.filter(
            ~models.StorageGateway.deleted,
            models.StorageGateway.display_name == display_name,
        ).one_or_none()

    def _create_storage_gateway(self, doc):
        with current_app.appsync_client.pause_notifications():
            current_app.logger.debug(f"Creating storage gateway: {doc}")
            try:
                cls = models.StorageGateway.polymorphic_class(doc["connector_id"])
                gateway = cls(doc=doc)
            except (TypeError, ValueError) as e:
                click.echo(
                    "Error creating storage gateway model: {}".format(str(e)), err=True
                )
                click.get_current_context().exit(1)
            create_object(gateway)
        return gateway

    def _update_storage_gateway(self, update_doc, model):
        current_app.logger.debug(f"Updating storage gateway {update_doc}")
        with current_app.appsync_client.pause_notifications():
            patch_object(model, update_doc)

    def _create_gateway_update_doc(self, doc, model):
        """
        Create a doc to update the storage gateway based on differences between the
        model and doc. If no changes are detcted, we return {}. Limiting updates
        significantly speeds up succesive runs of migrate4.

        We only check the values that gcs_migration's gateway model provides.
        """
        update_doc = {}
        # Ignoring display_name since that's how we find the gateway
        for attr in [
            "allowed_domains",
            "users_allow",
            "users_deny",
        ]:
            org_value = getattr(model, attr, None)
            new_value = doc.get(attr)
            if (org_value is None) ^ (new_value is None) or (
                org_value is not None
                and new_value is not None
                and set(org_value) != set(new_value)
            ):
                update_doc[attr] = new_value

        model_value = getattr(model, "identity_mappings", None)
        doc_value = doc.get("identity_mappings")
        if self._cmp_identity_mappings(model_value, doc_value) != 0:
            update_doc["identity_mappings"] = doc_value

        model_value = getattr(model, "restrict_paths", None)
        doc_value = doc.get("restrict_paths")
        if self._cmp_restrict_paths(model_value, doc_value) != 0:
            update_doc["restrict_paths"] = doc_value

        for attr in ["groups_allow", "groups_deny"]:
            org_value = getattr(model, attr, None)
            new_value = doc["policies"].get(attr)
            if (org_value is None) ^ (new_value is None) or (
                org_value is not None
                and new_value is not None
                and set(org_value) != set(new_value)
            ):
                if "policies" not in update_doc:
                    update_doc["policies"] = {}
                update_doc["policies"][attr] = new_value

        # If there are changes, file out the boilerplate
        if update_doc:
            update_doc["DATA_TYPE"] = "storage_gateway#1.1.0"
            if update_doc.get("policies") is None:
                update_doc["policies"] = {}
            update_doc["policies"]["DATA_TYPE"] = "posix_storage_policies#1.0.0"

        return update_doc

    def _sync_mapped_collection(self, plan, storage_gateway):
        click.echo(
            f"Syncing v4 endpoint {plan.v4_collection_id} to v5 mapped collection"
        )

        # _sync_collection() will create the collection and the roles
        return self._sync_collection(
            plan, storage_gateway, authorized_identity={"id": plan.doc["identity_id"]},
        )

    def _sync_guest_collections(self, plans, storage_gateway, mapped_collection):
        # Phase I: Create/update missing guest collections
        for guest_plan in plans:
            if not guest_plan.disable:
                self._sync_guest_collection(
                    guest_plan, storage_gateway, mapped_collection
                )

        # Phase 2: Delete extraneous guest collections
        for guest in self._lookup_all_guest_collections(mapped_collection.id):
            if not self._is_guest_in_plans(plans, guest):
                click.echo(f"Deleting extraneous v5 guest collection {guest.id}")
                self._delete_guest_collection(guest)

    def _sync_guest_collection(self, plan, storage_gateway, mapped_collection):
        click.echo(f"Syncing v4 share {plan.v4_collection_id} to v5 guest collection")

        # Create this collection's user credential
        user_credental = self._sync_user_credential(plan.credential, storage_gateway)

        # _sync_collection() will create the collection and the roles
        collection = self._sync_collection(
            plan,
            storage_gateway,
            authorized_identity={
                "id": plan.doc["identity_id"],
                "username": plan.credential.identity_username,
                "status": "used",
            },
            mapped_collection=mapped_collection,
            user_credential=user_credental,
        )

        # Sync this collection's ACLs
        self._sync_acls(plan.acls, collection)
        return collection

    def _sync_collection(
        self,
        plan,
        storage_gateway,
        authorized_identity,
        mapped_collection=None,
        user_credential=None,
    ):
        """
        Create or update the collection

        We create the collection with a temporary ID. Later when the admin is ready, we'll
        provide the ability to switch the collection ID to the ID listed in the migration plan.

        We'll format the collection display name in order to avoid conflicts with the production
        v4 instance. We add a special keyword to the collection so that we can find the collection
        on successive iterations of this command.
        """
        # Check if the collection already exists
        current_app.logger.debug(
            f"Checking if collection '{plan.v4_collection_id}' exists"
        )

        collection, plan.v5_collection_id = self._lookup_collection(
            plan.doc["collection_type"],
            plan.v4_collection_id,
            storage_gateway.id,
            mapped_collection.id if mapped_collection else None,
            user_credential_id=user_credential.id if user_credential else None,
        )

        if not collection:
            click.echo(".Creating the new v5 collection")
            collection = self._create_collection(
                plan.doc,
                plan.v4_collection_id,
                storage_gateway,
                authorized_identity,
                mapped_collection,
                user_credential,
            )
            plan.v5_collection_id = collection.id
        elif not collection.id == plan.v4_collection_id:
            # Update collection if it hasn't been migrated already
            update_doc = self._create_collection_update_doc(
                plan.doc, plan.v4_collection_id, collection
            )
            if update_doc:
                click.echo(".Updating the v5 collection")
                self._update_collection(update_doc, collection)

        # Sync this collection's roles, but only if the collection hasn't been
        # migrated.
        if collection.id != plan.v4_collection_id:
            self._sync_roles(plan.roles, collection)
        return collection

    def _lookup_collection(
        self,
        collection_type,
        v4_collection_id,
        storage_gateway_id,
        mapped_collection_id=None,
        user_credential_id=None,
    ):
        """
        Search for a previously created collection by keyword.

        During migration, we create collections with a prefix that makes them more
        easily identifiable and to alert users that these are not the collections
        they are looking for. But there are a couple concerns:
         - collection names are not unique, so searching solo on that will break
         - we can't trust users to not change guest collection display_names

        So in addition to the display_name prefix, we add a keyword to help in both
        cases: "migration_<v4_collection_id>". It's not bulletproof but should be
        sufficient.
        """
        query = models.Collection.query.filter(
            ~models.Collection.deleted,
            models.Collection.collection_type == collection_type,
        )

        if mapped_collection_id:
            query = query.filter(
                models.Collection.mapped_collection_id == mapped_collection_id
            )
        else:
            query = query.filter(
                models.Collection.storage_gateway_id == storage_gateway_id
            )

        # Filter based on user credential ID to support cases where the migration plan
        # has changed the owner of the guest collection on successive runs.
        if user_credential_id:
            query = query.filter(
                models.Collection.user_credential_id == user_credential_id
            )

        matching_keyword = self._migration_keyword(v4_collection_id)

        # Filter collections by matching keyword
        def _keyword_search(col):
            for kw in getattr(col, "keywords", []):
                if kw == matching_keyword:
                    return True
            return False

        col_seq = filter(_keyword_search, query)
        collections = list(col_seq)

        if len(collections) > 1:
            # During migration, the v5 collection remains in the db until things
            # are complete. We do want to use the migrated collection however, so
            # that we don't try to redo migration steps that have already
            # completed
            migrated_collection = None
            v5_collection = None
            for c in collections:
                if c.id == v4_collection_id:
                    migrated_collection = c
                else:
                    v5_collection = c
            return migrated_collection, v5_collection.id
        elif collections:
            return collections[0], collections[0].id
        else:
            return None, None

    def _lookup_all_guest_collections(self, mapped_collection_id):
        return models.Collection.query.filter(
            models.Collection.collection_type == CollectionType.guest,
            ~models.UserCredential.deleted,
            models.Collection.mapped_collection_id.in_(
                [
                    mapped_collection_id,
                    self.plan_docs.mapped_collection.v5_collection_id,
                ]
            ),
        ).all()

    def _is_guest_in_plans(self, guest_plans, guest):
        """
        Check if the existing guest model is defined in the guest plans.
        """
        for keyword in getattr(guest, "keywords", []):
            if keyword.startswith(self.COLLECTION_KEYWORD_PREFIX):
                collection_id = keyword[len(self.COLLECTION_KEYWORD_PREFIX) :]
                for plan in guest_plans:
                    if (
                        collection_id == plan.v4_collection_id
                        and str(plan.credential.doc["identity_id"])
                        == guest.user_credential.identity_id
                        and plan.credential.doc["username"]
                        == guest.user_credential.username
                        and not plan.disable
                    ):
                        return True
        return False

    def _create_collection(
        self,
        doc,
        v4_collection_id,
        storage_gateway,
        authorized_identity,
        mapped_collection,
        user_credential,
    ):
        current_app.logger.debug(f"Creating collection: {doc}")

        modified_display_name = self._migration_display_name(doc["display_name"])
        modified_keywords = self._insert_migration_keyword(
            doc.get("keywords", []), v4_collection_id
        )

        create_doc = {
            **doc,
            "display_name": modified_display_name,
            "keywords": modified_keywords,
            "collection_type": doc["collection_type"],
        }
        if mapped_collection:
            create_doc["mapped_collection_id"] = mapped_collection.id

        with current_app.appsync_client.pause_notifications():
            try:
                domain_name = allocate_domain(doc["collection_type"])
                collection = models.Collection.from_collections_post(
                    storage_gateway=storage_gateway,
                    mapped_collection=mapped_collection,
                    endpoint=models.Endpoint.query.one(),
                    user_credential=user_credential,
                    domain_name=domain_name,
                    authorized_identity=authorized_identity,
                    doc=create_doc,
                )
            except Exception as e:
                click.echo(
                    f"Error creating collection: {str(e)}", err=True,
                )
                click.get_current_context().exit(1)
            create_object(collection)
            # If the endpoint is managed, there will be an owner role associated with
            # the collection. If we don't commit it here, it'll be in the DB but not
            # in appsync. migrate4 won't try to fix it later because it'll see it in
            # the DB.
            if collection.owner_role:
                create_object(collection.owner_role)
        return collection

    def _update_collection(self, update_doc, model):
        current_app.logger.debug(f"Updating collection {update_doc}")
        with current_app.appsync_client.pause_notifications():
            patch_object(model, update_doc)

    def _create_collection_update_doc(self, doc, v4_collection_id, model):
        """
        Create a doc to update the collection based on differences between the model and doc.
        If no changes are detcted, we return {}. Limiting updates significantly speeds up
        succesive runs of migrate4.

        We only check the values that gcs_migration's collection model provides.
        """
        update_doc = {}

        #
        # Common endpoint attributes
        #

        # display_name comparison
        modified_display_name = self._migration_display_name(doc["display_name"])
        if model.display_name != modified_display_name:
            update_doc["display_name"] = modified_display_name

        for attr in [
            "description",
            "info_link",
            "contact_info",
            "contact_email",
            "organization",
            "department",
        ]:
            if getattr(model, attr, None) != doc.get(attr):
                update_doc[attr] = doc.get(attr)

        # keywords comparison
        modified_keywords = self._insert_migration_keyword(
            doc.get("keywords", []), v4_collection_id
        )
        if set(getattr(model, "keywords", [])) != set(modified_keywords):
            update_doc["keywords"] = modified_keywords

        #
        # Mapped collection attributes
        #

        # Skipping these attributes:
        # - collection_base_path: immutable
        # - identity_id: immutable
        if model.collection_type == CollectionType.mapped.value:
            for attr, default in [
                ("default_directory", None),
                ("public", None),
                ("force_encryption", None),
                ("force_verify", None),
                ("disable_verify", None),
                ("allow_guest_collections", None),
                ("enable_https", True),
                ("user_message", None),
                ("user_message_link", None),
            ]:
                if getattr(model, attr, None) != doc.get(attr, default):
                    update_doc[attr] = doc.get(attr, default)

            for attr in ["sharing_users_allow", "sharing_users_deny"]:
                org_value = getattr(model, attr, None)
                new_value = doc.get(attr)
                if (org_value is None) ^ (new_value is None) or (
                    org_value is not None
                    and new_value is not None
                    and set(org_value) != set(new_value)
                ):
                    update_doc[attr] = new_value

            for attr in ["sharing_groups_allow", "sharing_groups_deny"]:
                org_value = getattr(model, attr, None)
                new_value = doc["policies"].get(attr)
                if (org_value is None) ^ (new_value is None) or (
                    org_value is not None
                    and new_value is not None
                    and set(org_value) != set(new_value)
                ):
                    if "policies" not in update_doc:
                        update_doc["policies"] = {}
                    update_doc["policies"][attr] = new_value

            model_value = model.sharing_restrict_paths
            doc_value = doc.get("sharing_restrict_paths")
            if self._cmp_sharing_restrict_paths(model_value, doc_value) != 0:
                update_doc["sharing_restrict_paths"] = doc.get("sharing_restrict_paths")

        #
        # Guest collection attributes
        #

        # Skipping these attributes:
        # - collection_base_path: immutable
        # - identity_id: immutable
        # - mapped_collection_id: immutable
        if model.collection_type == CollectionType.guest.value:
            for attr, default in [
                ("default_directory", None),
                ("public", None),
                ("force_encryption", False),
                ("force_verify", False),
                ("disable_verify", False),
                ("enable_https", True),
            ]:
                if getattr(model, attr, None) != doc.get(attr, default):
                    update_doc[attr] = doc.get(attr, default)

        if update_doc:
            update_doc["DATA_TYPE"] = "collection#1.4.0"
            if "policies" not in update_doc:
                update_doc["policies"] = {}
            update_doc["policies"]["DATA_TYPE"] = "posix_collection_policies#1.1.0"
        return update_doc

    def _delete_guest_collection(self, guest):
        with current_app.appsync_client.pause_notifications():
            # (soft) Delete the collection
            delete_object(guest, update=True)

            # Delete appsync representations of the role objects
            # for the collection
            for r in models.Role.query.filter(models.Role.resource == guest.id):
                if r.primary_role:
                    r.primary_role = False
                    r.refresh_configuration_hash()
                delete_object(r, collection_deleted=True)
            delete_object(guest, update=False)

    def _sync_user_credential(self, plan, storage_gateway):
        """
        Sync the user credentials
        """
        click.echo(".Syncing the user credential")

        # Check if the user credential already exists
        current_app.logger.debug(f"Checking if user credential exists: {plan.doc}")

        cred = models.UserCredential.query.filter(
            ~models.UserCredential.deleted,
            models.UserCredential.storage_gateway_id == storage_gateway.id,
            models.UserCredential.identity_id == str(plan.doc["identity_id"]),
            models.UserCredential.username == plan.doc["username"],
        ).one_or_none()

        if cred:
            return cred

        current_app.logger.debug("Creating user credential")
        click.echo("..Creating the new user credential")
        try:
            cls = models.UserCredential.polymorphic_class(str(plan.doc["connector_id"]))
            doc = {
                **plan.doc,
                "storage_gateway_id": str(storage_gateway.id),
                "identity_id": str(plan.doc["identity_id"]),
                "connector_id": str(plan.doc["connector_id"]),
            }
            cred = cls(doc=doc, storage_gateway=storage_gateway)
        except (TypeError, ValueError) as e:
            click.echo(
                "Error creating user credential model: {}".format(str(e)), err=True
            )
            click.get_current_context().exit(1)
        create_object(cred)
        return cred

    def _sync_acls(self, acl_plans, collection):
        click.echo(f".Syncing ACLs on v5 guest collection {collection.id}")

        existing_acls = self._lookup_all_acls(collection.id)

        # Phase I: Create missing ACLs
        for plan in acl_plans:
            acl = self._lookup_acl(
                plan.doc["principal"],
                plan.doc["principal_type"],
                plan.doc["path"],
                existing_acls,
            )
            if not acl:
                click.echo(f"..Creating v5 ACL for v4 ACL {plan.v4_acl_id}")
                self._create_acl(plan.doc, collection.id)
            elif plan.doc["permissions"] != acl["permissions"]:
                click.echo(f"..Updating v5 ACL {acl['id']}")
                self._update_acl(plan.doc, acl, collection.id)

        # Phase II: Delete extraneous ACLs
        for acl in existing_acls:
            # ACLs related to roles do not have IDs
            if acl["id"] and not self._is_acl_in_plans(acl_plans, acl):
                click.echo(f"..Deleting extraneous v5 ACL {acl['id']}")
                self._delete_acl(acl["id"], collection.id)

    def _lookup_acl(self, principal, principal_type, path, existing_acls):
        for acl in existing_acls:
            if (
                acl["principal"] == principal
                and acl["principal_type"] == principal_type
                and acl["path"] == path
            ):
                return acl
        return None

    def _lookup_all_acls(self, collection_id):
        """
        Fetch all ACLs for the collection
        """
        try:
            transfer_client = current_app.transfer_client
            res = transfer_client.endpoint_acl_list(collection_id)
        except globus_sdk.TransferAPIError as e:
            click.echo(
                f"Could not get ACLs for guest collection {collection_id}: {e}",
                err=True,
            )
            click.get_current_context().exit(1)
        return res["DATA"]

    def _is_acl_in_plans(self, acl_plans, acl):
        for plan in acl_plans:
            if (
                acl["principal"] == plan.doc["principal"]
                and acl["principal_type"] == plan.doc["principal_type"]
                and acl["path"] == plan.doc["path"]
            ):
                return True
        return False

    def _create_acl(self, doc, collection_id):
        try:
            transfer_client = current_app.transfer_client
            transfer_client.add_endpoint_acl_rule(collection_id, doc)
        except globus_sdk.TransferAPIError as e:
            click.echo(
                f"Could not add ACL to guest collection {collection_id}: {e}", err=True,
            )
            click.get_current_context().exit(1)

    def _update_acl(self, doc, acl, collection_id):
        try:
            transfer_client = current_app.transfer_client
            transfer_client.update_endpoint_acl_rule(collection_id, acl["id"], doc)
        except globus_sdk.TransferAPIError as e:
            click.echo(
                f"Could not update ACL on guest collection {collection_id}: {e}",
                err=True,
            )
            click.get_current_context().exit(1)

    def _delete_acl(self, acl_id, collection_id):
        try:
            transfer_client = current_app.transfer_client
            transfer_client.delete_endpoint_acl_rule(collection_id, acl_id)
        except globus_sdk.TransferAPIError as e:
            click.echo(
                f"Could not delete ACL {acl_id} on guest collection {collection_id}: {e}",
                err=True,
            )
            click.get_current_context().exit(1)

    def _sync_roles(self, role_plans, collection):
        click.echo(
            f".Syncing roles on v5 {collection.collection_type} collection {collection.id}"
        )

        # Phase I: Create missing roles
        for plan in role_plans:
            role = self._lookup_role(
                plan.doc["principal"], plan.doc["role"], collection.id
            )
            if not role:
                click.echo(f"..Creating v5 role for v4 role {plan.v4_role_id}")
                self._create_role(plan.doc, collection.id)
            # We can not update roles, just delete them

        # Phase II: Delete extraneous roles
        for role in self._lookup_all_roles(collection.id):
            if not self._is_role_in_plans(role_plans, role) and not role.primary_role:
                click.echo(f"..Deleting extraneous v5 role {role.id}")
                self._delete_role(role)

    def _lookup_role(self, principal, role, collection_id):
        """
        Lookup a single role in the database.
        """
        return models.Role.query.filter(
            ~models.Role.deleted,  # type: ignore
            models.Role.principal == principal,
            models.Role.role_name == role,
            models.Role.resource.is_(collection_id),  # type: ignore
        ).one_or_none()

    def _lookup_all_roles(self, collection_id):
        """
        Lookup all roles for a collection in the database.
        """
        return models.Role.query.filter(
            ~models.Role.deleted,  # type: ignore
            models.Role.resource.is_(collection_id),  # type: ignore
        ).all()

    def _is_role_in_plans(self, role_plans, role):
        for plan in role_plans:
            plan_role_type = RoleType("collection:" + plan.doc["role"])
            if plan.doc["principal"] == role.principal and plan_role_type == role.role:
                return True
        return False

    def _create_role(self, doc, collection_id):
        with current_app.appsync_client.pause_notifications():
            role = models.Role(doc={**doc, "collection": collection_id})
            create_object(role)
        return role

    def _delete_role(self, role):
        if role.primary_role:
            click.echo(
                f"Cannot delete primary role {role.id} on collection {role.resource}",
                err=True,
            )
            return
        with current_app.appsync_client.pause_notifications():
            delete_object(role)

    def _check_globus_environment(self, app_env, plan_env):
        if plan_env == "default":
            plan_env = "production"

        if plan_env is None:
            if app_env in ["default", "production"]:
                return
            click.echo(
                "The migration plan was created for the production environment, it will not work with "
                f"an endpoint in the {app_env} environment.",
                err=True,
            )
            click.get_current_context().exit(1)

        if app_env.upper() != plan_env.upper():
            click.echo(
                f"The migration plan was created for the {plan_env} environment, it will not work with "
                f"an endpoint in the {app_env} environment.",
                err=True,
            )
            click.get_current_context().exit(1)

    def _migration_display_name(self, display_name):
        return self.COLLECTION_DISPLAY_NAME_FORMAT.format(display_name)

    def _migration_keyword(self, v4_collection_id):
        return self.COLLECTION_KEYWORD_FORMAT.format(v4_collection_id)

    def _insert_migration_keyword(self, keywords, v4_collection_id):
        return keywords + [self._migration_keyword(v4_collection_id)]

    # Used throughout the following comparison functions
    MATCH = 0
    NO_MATCH = 1

    def _cmp_path_restrictions(self, model_value, doc_value):
        """
        Compare path_restrictions between the existing model and the migration doc.
        Returns 0 if values match, non-zero otherwise.
        """
        for field in ["read", "read_write", "none"]:
            if set(getattr(model_value, field)) != set(doc_value.get(field, [])):
                return self.NO_MATCH
        return self.MATCH

    def _cmp_sharing_restrict_paths(self, model_value, doc_value):
        """
        Compare mapped collection sharing_restrict_paths between the existing model
        and the migration doc. Returns 0 if values match, non-zero otherwise.

        The default sharing_restrict_paths value for the model is:
            {'_read': [], '_read_write': ['/'], '_none': []}
        """
        default_sharing_restrict_paths = {"read": [], "read_write": ["/"], "none": []}
        if doc_value is not None:
            return self._cmp_path_restrictions(model_value, doc_value)
        return self._cmp_path_restrictions(model_value, default_sharing_restrict_paths)

    def _cmp_restrict_paths(self, model_value, doc_value):
        """
        Compare storage gateway restrict_paths between the existing model
        and the migration doc. Returns 0 if values match, non-zero otherwise.

        The default restrict_paths value for the model is 'None'.
        """
        if model_value is None and doc_value is None:
            return self.MATCH
        if model_value is None or doc_value is None:
            return self.NO_MATCH
        return self._cmp_path_restrictions(model_value, doc_value)

    def _cmp_mapping_loop(self, outer_list, inner_list, cmp_func):
        """
        Try to find every item in outer_list in inner_list.
        """
        for outer_val in outer_list:
            found_match = False
            for inner_val in inner_list:
                if cmp_func(outer_val, inner_val) == self.MATCH:
                    found_match = True
                    break
            if not found_match:
                return self.NO_MATCH
        return self.MATCH

    def _cmp_expression_mappings(self, model_value, doc_value):
        def _cmp_expression_mapping(model_mapping, doc_mapping):
            for attr, default in [
                ("source", None),
                ("match", None),
                ("ignore_case", False),
                ("literal", False),
                ("output", None),
            ]:
                if model_mapping.get(attr, default) != doc_mapping.get(attr, default):
                    return self.NO_MATCH
            return self.MATCH

        rv = self._cmp_mapping_loop(
            model_value.mappings, doc_value["mappings"], _cmp_expression_mapping,
        )
        if rv == self.NO_MATCH:
            return self.NO_MATCH

        return self._cmp_mapping_loop(
            doc_value["mappings"], model_value.mappings, _cmp_expression_mapping,
        )

    def _cmp_external_mapping(self, model_value, doc_value):
        if model_value.command != doc_value["command"]:
            return self.NO_MATCH
        return self.MATCH

    def _cmp_identity_mappings(self, model_value, doc_value):
        """
        Compare the identity_mappings on the storage gateway to the mappings supplied in
        the migration plan. Return MATCH if the two are equivalent, return NO_MATCH otherwise.
        """
        if model_value is None and doc_value is None:
            return self.MATCH
        if model_value is None or doc_value is None:
            return self.NO_MATCH

        def _cmp_identity_mapping(model_mapping, doc_mapping):
            """
            Compare two mappings which might not have the same DATA_TYPE.
            """
            if model_mapping.DATA_TYPE != doc_mapping["DATA_TYPE"]:
                return self.NO_MATCH
            mapping_type = model_mapping.DATA_TYPE.split("#")[0]
            if (
                mapping_type == "expression_identity_mapping"
                and self._cmp_expression_mappings(model_mapping, doc_mapping)
                == self.MATCH
            ):
                return self.MATCH

            if (
                mapping_type == "external_identity_mapping"
                and self._cmp_external_mapping(model_mapping, doc_mapping) == self.MATCH
            ):
                return self.MATCH

            return self.NO_MATCH

        rv = self._cmp_mapping_loop(model_value, doc_value, _cmp_identity_mapping,)
        if rv == self.NO_MATCH:
            return self.NO_MATCH

        return self._cmp_mapping_loop(
            doc_value, model_value, lambda x, y: _cmp_identity_mapping(y, x)
        )


def _import_migration_plan():
    """
    Load the GCS migration plan from disk.
    """
    plan_sub_directory = "migration_plan"

    if not os.path.isdir(plan_sub_directory):
        click.echo(
            f"{plan_sub_directory} not found in the current directory. "
            "Change to the directory containing the plan and rerun this command.",
            err=True,
        )
        click.get_current_context().exit(1)

    try:
        migration_plan = MigrationPlan.load()
    except KeyError as e:
        click.echo(
            f"Failed to load the migration plan because of an internal issue. "
            f'Please report this error to support@globus.org: \n"KeyError: {e}"',
            err=True,
        )
        click.get_current_context().exit(1)

    return migration_plan


def _check_imported_plan(migration_plan):
    """
    Run gcs_migration checks on on all components in the plan.
    """

    def _walk_plan(migration_plan):
        yield (check.MigrationPlanCheck, migration_plan)
        yield (check.StorageGatewayCheck, migration_plan.storage_gateway)
        yield (check.MappedCollectionCheck, migration_plan.mapped_collection)

        for role in migration_plan.mapped_collection.roles:
            yield (check.RoleCheck, role)

        for guest in migration_plan.guest_collections or []:
            # XXX skipping, see CH9351
            # yield (check.GuestCollectionCheck, guest)
            yield (check.CredentialCheck, guest.credential)

            for role in guest.roles:
                yield (check.RoleCheck, role)

            if guest.acls is not None:
                for acl in guest.acls:
                    yield (check.ACLCheck, acl)

    for t in _walk_plan(migration_plan):
        res = t[0]().check(t[1])
        if res:
            click.echo(
                "The migration plan is not ready. Run 'gcs-migration check' and fix all issues.",
                err=True,
            )
            click.get_current_context().exit(1)


def _load_doc(name, model, schema):
    """
    Helper function that wraps these calls in order to print useful information
    on exception.

    Note: Importing the the to_json() docs via our schema will transform some fields
    to make them suitable for use with our models. For example. collection.collection_type
    transforms from a string ('mapped', 'guest') to an enum.
    """
    try:
        doc_in = model.to_json()
    except AttributeError as e:
        click.echo(
            f"Failed to dump {name} document:\n"
            f"{str(e)}\n"
            "This is an internal error that needs to be reported to support@globus.org",
            err=True,
        )
        click.get_current_context().exit(1)

    try:
        if schema:
            doc_out = schema.load(doc_in)
        else:
            doc_out = doc_in
    except ma.ValidationError as e:
        click.echo(
            f"Failed to load {name} document:\n"
            f"{str(e)}\n"
            "This is an internal error that needs to be reported to support@globus.org",
            err=True,
        )
        click.get_current_context().exit(1)

    return doc_out


class MigrationPlanDocs:
    """
    This class creates a derivative of gcs_migration.MigrationPlan for use
    throughout this command. It is the schema-validated version of the to_json()
    calls from each component of MigrationPlan, plus additional information that
    is needed for the migration.

    The class is structured very similarily to MigrationPlan with the exception
    that the schema-validated doc is stored at the 'doc' attribute of each component.
    So it looks like:

    MigrationPlanDocs:
    - storage_gateway:
      - doc: {...}
      mapped_collection:
      - v4_collection_id
      - doc: {...}
      - roles[]
        - doc
      guest_collections[]:
      - v4_collection_id
      - doc: {...}
      - roles[]
        - v4_role_id
        - doc
      - acls[]
        - v4_acl_id
        - doc: {...}
      - credential
        - identity_username
        - doc: {...}
    """

    class StorageGateway:
        def __init__(self, imported_gateway_plan):
            self.doc = _load_doc(
                "storage gateway", imported_gateway_plan, schema.storage_gateway_schema,
            )

    class MappedCollection:
        def __init__(self, imported_collection_plan, role_docs):
            # Save off the collection id for the final migration step
            self.v4_collection_id = imported_collection_plan.id
            self.v5_collection_id = None
            self.doc = _load_doc(
                "mapped collection", imported_collection_plan, schema.collection_schema,
            )

            self.roles = role_docs

    class GuestCollection:
        def __init__(
            self, imported_guest_collection, cred_doc, role_docs, acl_docs, disable
        ):
            # Save off the collection id for the final migration step
            self.v4_collection_id = imported_guest_collection.id
            self.v5_collection_id = None
            self.disable = disable
            if disable:
                self.doc = None
            else:
                self.doc = _load_doc(
                    "guest collection",
                    imported_guest_collection,
                    schema.collection_schema,
                )

            self.credential = cred_doc
            self.roles = role_docs
            self.acls = acl_docs

    class Role:
        def __init__(self, imported_role):
            self.v4_role_id = imported_role.id
            self.doc = _load_doc("role", imported_role, schema.role_schema)

    class ACL:
        def __init__(self, imported_acl):
            self.v4_acl_id = imported_acl.id
            self.doc = _load_doc("ACL", imported_acl, None)

    class Credential:
        def __init__(self, imported_credential):
            self.identity_username = imported_credential.identity_username
            # Inject a storage_gateway_id in order to pass the schema checks
            imported_credential.storage_gateway_id = (
                "DEADBEEF-DEAD-BEEF-DEAD-BEEFDEADBEAF"
            )
            self.doc = _load_doc(
                "credential", imported_credential, schema.user_credential_schema
            )

    def __init__(self, imported_plan):
        self.storage_gateway = self.StorageGateway(imported_plan.storage_gateway)

        role_docs = [self.Role(role) for role in imported_plan.mapped_collection.roles]
        self.mapped_collection = self.MappedCollection(
            imported_plan.mapped_collection, role_docs
        )

        self.guest_collections = []
        for guest in imported_plan.guest_collections or []:
            # Credential
            cred_doc = self.Credential(guest.credential)
            # Roles
            role_docs = [self.Role(role) for role in guest.roles]
            # ACLs
            acls = guest.acls if guest.acls is not None else []
            acl_docs = [self.ACL(acl) for acl in acls]

            self.guest_collections.append(
                self.GuestCollection(
                    guest, cred_doc, role_docs, acl_docs, guest.disable
                )
            )
