"""
globus-connect-server node create

The globus-connect-server node create command creates a node UUID and adds
a node key to the GCS endpoint's keychain. It outputs a node configuration
that can be used as an argument to the --import-node command-line option of
globus-connect-server node setup
"""
import json
import os
import uuid

import click

from gcs_cli.parsing.command_decorator import globus_command

import globus.manager.api.models as models
from globus.manager.constants import ConfigurationObjectTag
from globus.manager.appsync.query import GraphQLError
from globus.manager.api.app import make_app
from globus.manager.utils.keychain import NodeKey


class NodeKeyFileParam(click.File):
    name = "node_key_file"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def convert(self, value, param, ctx):
        with super().convert(value, param, ctx) as f:
            try:
                node_key = NodeKey(**json.load(f))
            except Exception as e:
                self.fail("Error parsing {}: {}".format(value, str(e)), param, ctx)
        return node_key


@globus_command(
    "create",
    short_help="Create a Globus Connect Server node configuration",
    help="Create a Globus Connect Server node configuration",
    disable_options=["format", "explicit_host"],
)
@click.option(
    "--client-id",
    "-c",
    help="Auth Client ID for this Globus Connect Server Endpoint",
    required=True,
)
@click.option(
    "--secret",
    "-s",
    help="Client secret for the auth client",
    prompt=True,
    hide_input=True,
    envvar="GLOBUS_CLIENT_SECRET",
    show_envvar=True,
)
@click.option(
    "--deployment-key",
    help="Path to deployment key",
    default="deployment-key.json",
    type=NodeKeyFileParam(),
)
@click.option(
    "--export-node",
    type=click.Path(dir_okay=False, writable=True),
    required=True,
    metavar="NODE_INFO_JSON",
    help="File to write node configuration to for deployment.",
)
def node_create(client_id, secret, deployment_key, export_node):
    ctx = click.get_current_context()

    environment = os.environ.get("GLOBUS_SDK_ENVIRONMENT", "production")

    if not client_id:
        click.echo("Invalid --client-id", err=True)
        ctx.exit(1)
    if not secret:
        click.echo("Invalid secret", err=True)
        ctx.exit(1)

    app = make_app(app_config="globus.manager.api.config.endpoint_cli")
    models.db.init_app(app)
    app.logger.info("setup_node: start")
    app.config["DEPLOYMENT_KEY"] = deployment_key
    app.config["CLIENT_ID"] = client_id
    app.config["CLIENT_SECRET"] = secret
    app.config["ENVIRONMENT"] = environment
    app.keychain.add_node_key(deployment_key)

    with app.app_context():
        models.db.create_all()

        try:
            download_keychain(app)
            node_id = str(uuid.uuid4())
            node_key = create_node_key(app)
        except Exception as e:
            click.echo(str(e), err=True)
            ctx.exit(1)

        fd = os.open(export_node, os.O_WRONLY | os.O_TRUNC | os.O_CREAT, 0o600)
        with open(fd, "w") as f:
            json.dump(
                {"node_key": node_key.export(private_key=True), "node_id": node_id},
                f,
                indent=2,
            )
    app.logger.info("create_node: end")


def download_keychain(app):
    try:
        items = app.appsync_client.query_all(id=ConfigurationObjectTag.KEY_CHAIN.value)
        for item in items:
            app.appsync_client.process_item(item)
    except GraphQLError as e:
        raise Exception("Error downloading keychain: {}".format(e))


def create_node_key(app):
    return app.keychain.add_node_key()
