import os
import json
import time
from typing import Optional

import click
from flask import current_app
import requests

import globus.manager.api.models as models
from globus.manager.api.schemas.api.domain import DomainDict
from globus.manager.api.app import make_app
from gcs_cli.parsing.command_decorator import globus_command
from globus.manager.constants import NodeStatus
from globus.manager.scripts.utils.setup_utils import load_keychain
from globus.manager.scripts.utils.dns_patterns import (
    verify_hostname_option,
    verify_domain_name_option,
)
import globus.manager.utils.crypto_utils as crypto_utils
from globus.manager.utils.keychain import NodeKey


@globus_command(
    name="create",
    short_help="Create an OIDC server for this endpoint",
    disable_options=["format", "explicit_host"],
    help="""\
The OIDC create command creates an OIDC server configuration for this endpoint.

Once configured, install the OIDC server on each node by running
`globus-connect-server oidc setup`.

== EXAMPLE

globus-connect-server oidc create

""",
)
@click.option("--display-name", help="Display name for the OIDC server", required=True)
@click.option(
    "--support-contact", help="Support contact name for the OIDC server", required=True
)
@click.option(
    "--support-email",
    help="Support contact email address for the OIDC server",
    required=True,
)
@click.option(
    "--pam-service",
    "-p",
    help="Specify the PAM service module to use for authentication. Default is 'login'",
    default="login",
)
@click.option(
    "--server-name",
    metavar="PATH",
    help="Fully-qualified domain name for the OIDC service",
    callback=verify_domain_name_option,
)
@click.option(
    "--certificate-path",
    metavar="PATH",
    help="Path to the certificate for the virtual host for the OIDC service",
)
@click.option(
    "--certificate-chain-path",
    metavar="PATH",
    help="Path to the certificate chain for the virtual host for the OIDC service",
)
@click.option(
    "--private-key-path",
    metavar="PATH",
    help="Path to the private key for the virtual host for the OIDC service",
)
@click.option(
    "--manage-certificate-and-key",
    is_flag=True,
    default=False,
    help="""\
Encrypt and synchronize the certificate and key between data transfer nodes; if
not passed, you are responsible for ensuring that the certificate, chain, and
key files are present on each data transfer node
""",
)
@click.option(
    "--quickstart-server-name",
    help="""\
DNS label to use for configuring the OIDC server to use a subdomain of
the endpoint's domain.
""",
    callback=verify_hostname_option,
)
@click.pass_context
def oidc_create(
    ctx,
    display_name,
    support_contact,
    support_email,
    pam_service,
    server_name,
    certificate_path,
    certificate_chain_path,
    private_key_path,
    manage_certificate_and_key,
    quickstart_server_name,
) -> None:
    if os.getuid() != 0:
        click.echo("This command must be run as the root user", err=True)
        ctx.exit(1)

    app = make_app(app_config="globus.manager.api.config.oidc_cli")
    models.db.init_app(app)

    with app.app_context():
        models.db.create_all()
        models.db.session.commit()

        # Needed by AppSync
        for conf in models.Configuration.query:
            if conf.name == "node_key":
                value = NodeKey(**json.loads(conf.value))
            else:
                value = conf.value
            app.config[conf.name.upper()] = value

        # Make sure `node setup` has been run
        if "CLIENT_ID" not in app.config:
            click.echo("This node has not been set up.", err=True)
            click.get_current_context().exit(1)

        try:
            # Confirm that the loaded configuration contains the client ID and secret,
            # otherwise this node has not been configured yet. Can also be checked in app.py:266.
            if not (app.config.keys() > {"CLIENT_ID", "CLIENT_SECRET"}):
                raise Exception(
                    "GCS node has not been configured. Please run 'node setup' "
                    "or run this command on an existing node."
                )

            # Load the keychain
            load_keychain()

            # Get the endpoint
            endpoint_model = models.Endpoint.query.one()
            if endpoint_model.has_oidc_server and hasattr(
                endpoint_model.oidc_server, "auth_identity_provider_data"
            ):
                click.echo(
                    "There is already an OIDC configuration for this endpoint, "
                    "please use the `oidc update` command to update parameters."
                )
                return
            elif not endpoint_model.has_oidc_server:
                if quickstart_server_name is not None:
                    # For quickstart_server_name, the domain, cert, and key related
                    # options are disallowed
                    check_quickstart_arg_conflicts(
                        quickstart_server_name,
                        server_name,
                        certificate_path,
                        certificate_chain_path,
                        private_key_path,
                        manage_certificate_and_key,
                    )

                    # The OIDC server is given a subdomain of the endpoint
                    server_name = ".".join(
                        [quickstart_server_name, endpoint_model.collection_subdomain]
                    )

                    # Configure the OIDC server to use the same cert as the endpoint
                    if endpoint_model.has_domain:
                        domain = endpoint_model.domain
                    else:
                        domain = endpoint_model.certificate.domain
                    certificate_chain_path = domain.certificate_chain_path
                    private_key_path = domain.private_key_path
                    certificate_path = domain.certificate_path
                    certificate_chain_path = domain.certificate_chain_path
                    private_key_path = domain.private_key_path
                    manage_certificate_and_key = False

                domain = parse_domain_args(
                    server_name,
                    certificate_path,
                    certificate_chain_path,
                    private_key_path,
                    manage_certificate_and_key,
                )

                endpoint_model.oidc_server = models.OIDCServer(
                    doc={
                        "DATA_TYPE": "oidc_server#1.0.0",
                        "domain": domain,
                        "display_name": display_name,
                        "support_contact": support_contact,
                        "support_email": support_email,
                        "pam_service": pam_service,
                    }
                )
                endpoint_model.refresh_configuration_hash()
                current_app.appsync_client.update_object(endpoint_model)
                models.db.session.add(endpoint_model)
                models.db.session.commit()

            try:
                idp_data = endpoint_model.oidc_server.auth_identity_provider_data
                # FIXME: I can't tell if the idp data corresponds
                # to a registration for this domain for a different client ID
                # that got nuked
            except AttributeError:
                idp_data = None

            if not idp_data:
                # If the OIDC IDP hasn't been registered yet, we will try to do
                # so
                globus_dns_domain = models.GlobusDnsDomain.query.one_or_none()
                if globus_dns_domain:
                    any_active_node = False
                    for node in globus_dns_domain.nodes:
                        if node.status == NodeStatus.active:
                            any_active_node = True
                            break
                    if any_active_node:
                        resp = None
                        # Make sure the autodiscovery URL is reachable
                        for _tries in range(10):
                            time.sleep(1)
                            try:
                                resp = requests.get(
                                    endpoint_model.oidc_server.autodiscovery_url
                                )
                                if resp.status_code == 200:
                                    break
                            except Exception as e:
                                err = "Error contacting autodiscovery url: {}".format(
                                    str(e)
                                )
                                if _tries < 9:
                                    current_app.logger.info(err)
                                else:
                                    click.echo(err, err=True)
                                    click.get_current_context().exit(1)

                        if resp and resp.status_code == 200:
                            try:
                                endpoint_model.oidc_server.register_idp()
                            except Exception as e:
                                click.echo(
                                    "Error registering IdP: {}".format(str(e)), err=True
                                )
                                click.get_current_context().exit(1)

                            endpoint_model.refresh_configuration_hash()
                            current_app.appsync_client.update_object(endpoint_model)
                            click.echo("OIDC server enabled for this endpoint.")
                            return
                click.echo(
                    "The OIDC Server configuration has been created, but there "
                    "are currently no reachable data transfer nodes for this "
                    "endpoint. Please run globus-connect-server node setup to "
                    "complete the service installation."
                )
                click.get_current_context().exit(1)
        except Exception as e:
            click.echo("Error while creating OIDC endpoint: {}".format(str(e)))
            click.get_current_context().exit(1)


def check_quickstart_arg_conflicts(
    quickstart_server_name: str,
    server_name: Optional[str],
    certificate_path: Optional[str],
    certificate_chain_path: Optional[str],
    private_key_path: Optional[str],
    manage_certificate_and_key: bool,
) -> None:
    non_qs_options = {
        "--server-name": server_name,
        "--certificate-path": certificate_path,
        "--certificate-chain-path": certificate_chain_path,
        "--private-key-path": private_key_path,
        "--manage-certificate-and-key": manage_certificate_and_key,
    }
    for non_qs_option, val in non_qs_options.items():
        if val not in [None, False]:
            click.echo(
                "The --quickstart-server-name option can not be "
                "used with the {} option".format(non_qs_option)
            )
            click.get_current_context().exit(1)
    if "." in quickstart_server_name:
        click.echo("The --quickstart-server-name value must not contain '.'")
        click.get_current_context().exit(1)
    elif quickstart_server_name.startswith("m-") or quickstart_server_name.startswith(
        "g-"
    ):
        click.echo("The --quickstart-server-name value must not start with m- or g-")
        click.get_current_context().exit(1)


def parse_domain_args(
    hostname: str,
    certificate_path: Optional[str],
    certificate_chain_path: Optional[str],
    private_key_path: Optional[str],
    manage_certificate_and_key: bool,
) -> DomainDict:
    if certificate_path is None:
        raise click.ClickException("Missing --certificate-path")
    elif private_key_path is None:
        raise click.ClickException("Missing --private-key-path")

    if manage_certificate_and_key:
        certificate_chain: Optional[str] = None
        with open(certificate_path, "r") as f:
            data = f.read()
        certificate = crypto_utils.parse_certificate(data, hostname)
        if certificate_chain_path is not None:
            with open(certificate_chain_path, "r") as f:
                data = f.read()
            chain = crypto_utils.parse_certificate_chain(data)
            if chain is not None:
                certificate_chain = chain
        with open(private_key_path, "r") as f:
            data = f.read()
        private_key = crypto_utils.parse_private_key(data)

        return {
            "DATA_TYPE": "domain#1.0.0",
            "domain_name": hostname,
            "wildcard": False,
            "certificate": certificate,
            "certificate_chain": certificate_chain,
            "private_key": private_key,
        }
    else:
        return {
            "DATA_TYPE": "domain#1.0.0",
            "domain_name": hostname,
            "wildcard": False,
            "certificate_path": certificate_path,
            "certificate_chain_path": certificate_chain_path,
            "private_key_path": private_key_path,
        }
